# Copyright 2014 OpenMarket Ltd
# Copyright 2019 New Vector Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import binascii
import json
import logging
from abc import abstractmethod
from typing import TYPE_CHECKING, Dict, Generic, Optional, Sequence, TypeVar

import signedjson.key
import signedjson.sign
from twisted.internet import defer
from twisted.internet.defer import Deferred
from twisted.python.failure import Failure
from twisted.web.client import readBody
from twisted.web.iweb import IResponse
from unpaddedbase64 import decode_base64

from sydent.config.exceptions import ConfigError
from sydent.db.hashing_metadata import HashingMetadataStore
from sydent.db.threepid_associations import GlobalAssociationStore, SignedAssociations
from sydent.threepid import threePidAssocFromDict
from sydent.types import JsonDict
from sydent.util import json_decoder
from sydent.util.hash import sha256_and_url_safe_base64
from sydent.util.stringutils import normalise_address

PushUpdateReturn = TypeVar("PushUpdateReturn")

if TYPE_CHECKING:
    from sydent.sydent import Sydent

logger = logging.getLogger(__name__)

SIGNING_KEY_ALGORITHM = "ed25519"


class Peer(Generic[PushUpdateReturn]):
    def __init__(self, servername: str, pubkeys: Dict[str, str]):
        """
        :param server_name: The peer's server name.
        :param pubkeys: The peer's public keys in a Dict[key_id, key_b64]
        """
        self.servername = servername
        self.pubkeys = pubkeys
        self.is_being_pushed_to = False

    @abstractmethod
    def pushUpdates(self, sgAssocs: SignedAssociations) -> "Deferred[PushUpdateReturn]":
        """
        :param sgAssocs: Map from originId to sgAssoc,  where originId is the id
                         on the creating server and sgAssoc is the json object
                         of the signed association
        """
        ...


class LocalPeer(Peer[bool]):
    """
    The local peer (ourselves: essentially copying from the local associations table to the global one)
    """

    def __init__(self, sydent: "Sydent") -> None:
        super().__init__(sydent.config.general.server_name, {})
        self.sydent = sydent
        self.hashing_store = HashingMetadataStore(sydent)

        globalAssocStore = GlobalAssociationStore(self.sydent)
        lastId = globalAssocStore.lastIdFromServer(self.servername)
        self.lastId = lastId if lastId is not None else -1

    def pushUpdates(self, sgAssocs: SignedAssociations) -> "Deferred[bool]":
        """
        Saves the given associations in the global associations store. Only stores an
        association if its ID is greater than the last seen ID.

        :param sgAssocs: The associations to save.

        :return: A deferred that succeeds with the value `True`.
        """
        globalAssocStore = GlobalAssociationStore(self.sydent)
        for localId in sgAssocs:
            if localId > self.lastId:
                assocObj = threePidAssocFromDict(sgAssocs[localId])

                # ensure we are casefolding email addresses
                assocObj.address = normalise_address(assocObj.address, assocObj.medium)

                if assocObj.mxid is not None:
                    # Assign a lookup_hash to this association
                    pepper = self.hashing_store.get_lookup_pepper()
                    if not pepper:
                        raise RuntimeError("No lookup_pepper in the database.")
                    str_to_hash = " ".join(
                        [
                            assocObj.address,
                            assocObj.medium,
                            pepper,
                        ],
                    )
                    assocObj.lookup_hash = sha256_and_url_safe_base64(str_to_hash)

                    # We can probably skip verification for the local peer (although it could
                    # be good as a sanity check)
                    globalAssocStore.addAssociation(
                        assocObj,
                        json.dumps(sgAssocs[localId]),
                        self.sydent.config.general.server_name,
                        localId,
                    )
                else:
                    globalAssocStore.removeAssociation(
                        assocObj.medium, assocObj.address
                    )

        d = defer.succeed(True)
        return d


class RemotePeer(Peer[IResponse]):
    def __init__(
        self,
        sydent: "Sydent",
        server_name: str,
        port: Optional[int],
        pubkeys: Dict[str, str],
        lastSentVersion: Optional[int],
    ) -> None:
        """
        :param sydent: The current Sydent instance.
        :param server_name: The peer's server name.
        :param port: The peer's port. Only used if no replication url is configured.
        :param pubkeys: The peer's public keys in a dict[key_id, key_b64]
        :param lastSentVersion: The ID of the last association sent to the peer.
        """
        super().__init__(server_name, pubkeys)
        self.sydent = sydent
        self.lastSentVersion = lastSentVersion

        # look up or build the replication URL
        replication_url = self.sydent.config.http.base_replication_urls.get(server_name)

        if replication_url is None:
            if not port:
                port = 1001
            replication_url = "https://%s:%i" % (server_name, port)

        if replication_url[-1:] != "/":
            replication_url += "/"

        # Capture the interesting bit of the url for logging.
        self.replication_url_origin = replication_url
        replication_url += "_matrix/identity/replicate/v1/push"
        self.replication_url = replication_url

        # Get verify key for this peer

        # Check if their key is base64 or hex encoded
        pubkey = self.pubkeys[SIGNING_KEY_ALGORITHM]
        try:
            # Check for hex encoding
            int(pubkey, 16)

            # Decode hex into bytes
            pubkey_decoded = binascii.unhexlify(pubkey)

            logger.warning(
                "Peer public key of %s is hex encoded. Please update to base64 encoding",
                server_name,
            )
        except ValueError:
            # Check for base64 encoding
            try:
                pubkey_decoded = decode_base64(pubkey)
            except Exception as e:
                raise ConfigError(
                    "Unable to decode public key for peer %s: %s" % (server_name, e),
                )

        self.verify_key = signedjson.key.decode_verify_key_bytes(
            SIGNING_KEY_ALGORITHM + ":", pubkey_decoded
        )

        # Attach metadata
        self.verify_key.alg = SIGNING_KEY_ALGORITHM
        self.verify_key.version = 0

    def verifySignedAssociation(self, assoc: JsonDict) -> None:
        """Verifies a signature on a signed association. Raises an exception if the
        signature is incorrect or couldn't be verified.

        :param assoc: A signed association.
        """
        if "signatures" not in assoc:
            raise NoSignaturesException()

        key_ids = signedjson.sign.signature_ids(assoc, self.servername)
        if (
            not key_ids
            or len(key_ids) == 0
            or not key_ids[0].startswith(SIGNING_KEY_ALGORITHM + ":")
        ):
            e = NoMatchingSignatureException(
                foundSigs=assoc["signatures"].keys(),
                requiredServername=self.servername,
            )
            raise e

        # Verify the JSON
        signedjson.sign.verify_signed_json(assoc, self.servername, self.verify_key)

    def pushUpdates(self, sgAssocs: SignedAssociations) -> "Deferred[IResponse]":
        """
        Pushes the given associations to the peer.

        :param sgAssocs: The associations to push.

        :return: A deferred which results in the response to the push request.
        """
        body = {"sgAssocs": sgAssocs}

        reqDeferred = self.sydent.replicationHttpsClient.postJson(
            self.replication_url, body
        )
        if reqDeferred is None:
            raise RuntimeError(f"Unable to push sgAssocs to {self.replication_url}")

        # XXX: We'll also need to prune the deleted associations out of the
        # local associations table once they've been replicated to all peers
        # (ie. remove the record we kept in order to propagate the deletion to
        # other peers).

        updateDeferred: "Deferred[IResponse]" = defer.Deferred()

        reqDeferred.addCallback(self._pushSuccess, updateDeferred=updateDeferred)
        reqDeferred.addErrback(self._pushFailed, updateDeferred=updateDeferred)

        return updateDeferred

    def _pushSuccess(
        self,
        result: "IResponse",
        updateDeferred: "Deferred[IResponse]",
    ) -> None:
        """
        Processes a successful push request. If the request resulted in a status code
        that's not a success, consider it a failure

        :param result: The HTTP response.
        :param updateDeferred: The deferred to make either succeed or fail depending on
            the status code.
        """
        if result.code >= 200 and result.code < 300:
            updateDeferred.callback(result)
        else:
            d = readBody(result)
            d.addCallback(self._failedPushBodyRead, updateDeferred=updateDeferred)
            d.addErrback(self._pushFailed, updateDeferred=updateDeferred)

    def _failedPushBodyRead(
        self, body: bytes, updateDeferred: "Deferred[IResponse]"
    ) -> None:
        """
        Processes a response body from a failed push request, then calls the error
        callback of the provided deferred.

        :param body: The response body.
        :param updateDeferred: The deferred to call the error callback of.
        """
        errObj = json_decoder.decode(body.decode("utf8"))
        e = RemotePeerError(errObj)
        updateDeferred.errback(e)

    def _pushFailed(
        self,
        failure: Failure,
        updateDeferred: "Deferred[object]",
    ) -> None:
        """
        Processes a failed push request, by calling the error callback of the given
        deferred with it.

        :param failure: The failure to process.
        :param updateDeferred: The deferred to call the error callback of.
        """
        updateDeferred.errback(failure)
        return None


class NoSignaturesException(Exception):
    pass


class NoMatchingSignatureException(Exception):
    def __init__(self, foundSigs: Sequence[str], requiredServername: str):
        self.foundSigs = foundSigs
        self.requiredServername = requiredServername

    def __str__(self) -> str:
        return "Found signatures: %s, required server name: %s" % (
            self.foundSigs,
            self.requiredServername,
        )


class RemotePeerError(Exception):
    def __init__(self, errorDict: JsonDict):
        self.errorDict = errorDict

    def __str__(self) -> str:
        return repr(self.errorDict)



# Copyright 2020 Dirk Klimpel
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING

from twisted.web.server import Request

from sydent.http.servlets import SydentResource, get_args, jsonwrap, send_cors
from sydent.types import JsonDict

if TYPE_CHECKING:
    from sydent.sydent import Sydent


class AuthenticatedUnbindThreePidServlet(SydentResource):
    """A servlet which allows a caller to unbind any 3pid they want from an mxid

    It is assumed that authentication happens out of band
    """

    def __init__(self, sydent: "Sydent") -> None:
        super().__init__()
        self.sydent = sydent

    @jsonwrap
    def render_POST(self, request: Request) -> JsonDict:
        send_cors(request)
        args = get_args(request, ("medium", "address", "mxid"))

        threepid = {"medium": args["medium"], "address": args["address"]}

        self.sydent.threepidBinder.removeBinding(
            threepid,
            args["mxid"],
        )
        return {}

    def render_OPTIONS(self, request: Request) -> bytes:
        send_cors(request)
        return b""



from typing import Dict

import attr
from typing_extensions import TypedDict

from sydent.types import JsonDict


class VerifyKey(TypedDict):
    key: str


VerifyKeys = Dict[str, VerifyKey]


@attr.s(frozen=True, slots=True, auto_attribs=True)
class CachedVerificationKeys:
    verify_keys: VerifyKeys
    valid_until_ts: int


# key: "signing key identifier"; value: signature encoded as unpadded base 64
# See https://spec.matrix.org/unstable/appendices/#signing-details
Signature = Dict[str, str]


@attr.s(frozen=True, slots=True, auto_attribs=True)
class SignedMatrixRequest:
    method: bytes
    uri: bytes
    destination_is: str
    signatures: Dict[str, Signature]
    origin: str
    content: JsonDict



# Copyright 2021 The Matrix.org Foundation C.I.C.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from configparser import ConfigParser
from typing import Optional

from sydent.config._base import BaseConfig


class HTTPConfig(BaseConfig):
    def parse_config(self, cfg: "ConfigParser") -> bool:
        """
        Parse the http section of the config

        :param cfg: the configuration to be parsed
        """
        # This option is deprecated
        self.verify_response_template = cfg.get(
            "http", "verify_response_template", fallback=None
        )

        self.client_bind_address = cfg.get("http", "clientapi.http.bind_address")
        self.client_port = cfg.getint("http", "clientapi.http.port")

        # internal port is allowed to be set to an empty string in the config
        internal_api_port = cfg.get("http", "internalapi.http.port")
        self.internal_bind_address = cfg.get(
            "http", "internalapi.http.bind_address", fallback="::1"
        )
        self.internal_port: Optional[int] = None
        if internal_api_port != "":
            self.internal_port = int(internal_api_port)

        self.cert_file = cfg.get("http", "replication.https.certfile")
        self.ca_cert_file = cfg.get("http", "replication.https.cacert")

        self.replication_bind_address = cfg.get(
            "http", "replication.https.bind_address"
        )
        self.replication_port = cfg.getint("http", "replication.https.port")

        self.obey_x_forwarded_for = cfg.getboolean("http", "obey_x_forwarded_for")

        self.verify_federation_certs = cfg.getboolean("http", "federation.verifycerts")

        self.server_http_url_base = cfg.get("http", "client_http_base")

        self.base_replication_urls = {}

        for section in cfg.sections():
            if section.startswith("peer."):
                # peer name is all the characters after 'peer.'
                peer = section[5:]
                if cfg.has_option(section, "base_replication_url"):
                    base_url = cfg.get(section, "base_replication_url")
                    self.base_replication_urls[peer] = base_url

        return False



#  Copyright 2022 The Matrix.org Foundation C.I.C.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
from http import HTTPStatus
from typing import Dict, Generic, Optional, TypeVar

from twisted.internet import task
from twisted.internet.interfaces import IReactorTime

from sydent.http.servlets import MatrixRestError

logger = logging.getLogger(__name__)

K = TypeVar("K")


class LimitExceededException(MatrixRestError):
    def __init__(self, error: Optional[str] = None) -> None:
        if error is None:
            error = "Too many requests"

        super().__init__(HTTPStatus.TOO_MANY_REQUESTS, "M_UNKNOWN", error)


class Ratelimiter(Generic[K]):
    """A ratelimiter based on leaky token bucket algorithm.

    Args:
        reactor
        burst: the number of requests that can happen at once before we start
            ratelimiting
        rate_hz: The maximum average sustained rate in hertz of requests we'll
            accept.
    """

    def __init__(self, reactor: IReactorTime, burst: int, rate_hz: float) -> None:
        # The "burst" count (or the capacity of each bucket in leaky bucket
        # algorithm).
        self._burst = burst

        # A map from key to number of tokens in its bucket. We ratelimit when
        # the number of tokens is greater than `burst`.
        #
        # Entries are removed when token count hits zero.
        self._buckets: Dict[K, int] = {}

        # We remove tokens from all buckets at `rate_hz` hertz.
        call = task.LoopingCall(self._periodic_call)
        call.clock = reactor
        call.start(1 / rate_hz)

    def _periodic_call(self) -> None:
        # Take one away from all active buckets. If a bucket reaches zero then
        # remove it from the dict.
        self._buckets = {
            key: tokens - 1 for key, tokens in self._buckets.items() if tokens > 1
        }

    def ratelimit(self, key: K, error: Optional[str] = None) -> None:
        """Check if we should ratelimit the request with the given key.

        Raises:
            LimitExceededException: if the request should be denied.
        """
        if error is None:
            error = "Too many requests"

        # We get the current token count and compare it with the `burst`.
        current_tokens = self._buckets.get(key, 0)
        if current_tokens >= self._burst:
            logger.warning("Ratelimit hit: %s: %s", error, key)
            raise LimitExceededException(error)

        self._buckets[key] = current_tokens + 1



# Copyright 2022 The Matrix.org Foundation C.I.C.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from twisted.test.proto_helpers import MemoryReactorClock
from twisted.trial import unittest

from sydent.util.ratelimiter import LimitExceededException, Ratelimiter


class RatelimiterTest(unittest.TestCase):
    def setUp(self) -> None:
        self.clock = MemoryReactorClock()
        self.ratelimiter = Ratelimiter(self.clock, burst=5, rate_hz=0.5)

    def test_simple(self) -> None:
        """Test that a request doesn't get ratelimited to start off with"""
        key = "key"

        # This should not raise as we're below the ratelimit
        self.ratelimiter.ratelimit(key)

    def test_burst(self) -> None:
        """Test that we can send `burst` number of messages before getting
        ratelimited
        """
        key = "key"

        # This should not raise as we're below the ratelimit
        for _ in range(5):
            self.ratelimiter.ratelimit(key)

        with self.assertRaises(LimitExceededException):
            self.ratelimiter.ratelimit(key)

    def test_burst_reset(self) -> None:
        """Test that once we hit the ratelimit we can wait a while and we'll be
        able to send requests again
        """
        key = "key"

        # This should not raise as we're below the ratelimit
        for _ in range(5):
            self.ratelimiter.ratelimit(key)

        with self.assertRaises(LimitExceededException):
            self.ratelimiter.ratelimit(key)

        self.clock.pump([2.0] * 5)

        for _ in range(5):
            self.ratelimiter.ratelimit(key)

        with self.assertRaises(LimitExceededException):
            self.ratelimiter.ratelimit(key)

    def test_average_rate(self):
        """Test that sending requests at a rate higher than the maximum rate
        gets ratelimited.
        """
        key = "key"

        with self.assertRaises(LimitExceededException):
            for _ in range(100):
                self.clock.advance(1)
                self.ratelimiter.ratelimit(key)

    def test_average_rate_burst(self):
        """Test that if we go above the maximum rate we'll get ratelimited"""
        key = "key"

        for _ in range(5):
            self.ratelimiter.ratelimit(key)

        for _ in range(100):
            self.clock.advance(2)
            self.ratelimiter.ratelimit(key)

        with self.assertRaises(LimitExceededException):
            self.ratelimiter.ratelimit(key)



# Copyright 2014 OpenMarket Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import urllib
from typing import TYPE_CHECKING, Dict, Optional

from sydent.db.valsession import ThreePidValSessionStore
from sydent.util import time_msec
from sydent.util.emailutils import sendEmail
from sydent.validators import common

if TYPE_CHECKING:
    from sydent.sydent import Sydent

logger = logging.getLogger(__name__)


class EmailValidator:
    def __init__(self, sydent: "Sydent") -> None:
        self.sydent = sydent

    def requestToken(
        self,
        emailAddress: str,
        clientSecret: str,
        sendAttempt: int,
        nextLink: Optional[str],
        ipaddress: Optional[str] = None,
        brand: Optional[str] = None,
    ) -> int:
        """
        Creates or retrieves a validation session and sends an email to the corresponding
        email address with a token to use to verify the association.

        :param emailAddress: The email address to send the email to.
        :param clientSecret: The client secret to use.
        :param sendAttempt: The current send attempt.
        :param nextLink: The link to redirect the user to once they have completed the
            validation.
        :param ipaddress: The requester's IP address.
        :param brand: A hint at a brand from the request.

        :return: The ID of the session created (or of the existing one if any)
        """
        valSessionStore = ThreePidValSessionStore(self.sydent)

        valSession, token_info = valSessionStore.getOrCreateTokenSession(
            medium="email", address=emailAddress, clientSecret=clientSecret
        )

        valSessionStore.setMtime(valSession.id, time_msec())

        # self.sydent.config.email.template is deprecated
        if self.sydent.config.email.template is None:
            templateFile = self.sydent.get_branded_template(
                brand,
                "verification_template.eml",
            )
        else:
            templateFile = self.sydent.config.email.template

        if token_info.send_attempt_number >= sendAttempt:
            logger.info(
                "Not mailing code because current send attempt (%d) is not less than given send attempt (%s)",
                sendAttempt,
                token_info.send_attempt_number,
            )
            return valSession.id

        ipstring = ipaddress if ipaddress else "an unknown location"

        substitutions = {
            "ipaddress": ipstring,
            "link": self.makeValidateLink(
                valSession.id, token_info.token, clientSecret, nextLink
            ),
            "token": token_info.token,
        }
        logger.info(
            "Attempting to mail code %s (nextLink: %s) to %s",
            token_info.token,
            nextLink,
            emailAddress,
        )
        sendEmail(self.sydent, templateFile, emailAddress, substitutions)

        valSessionStore.setSendAttemptNumber(valSession.id, sendAttempt)

        return valSession.id

    def makeValidateLink(
        self,
        session_id: int,
        token: str,
        clientSecret: str,
        nextLink: Optional[str],
    ) -> str:
        """
        Creates a validation link that can be sent via email to the user.

        :param session_id: The current validation session's ID.
        :param token: The token to make a link for.
        :param clientSecret: The client secret to include in the link.
        :param nextLink: The link to redirect the user to once they have completed the
            validation.

        :return: The validation link.
        """
        base = self.sydent.config.http.server_http_url_base
        link = "%s/_matrix/identity/api/v1/validate/email/submitToken?token=%s&client_secret=%s&sid=%d" % (
            base,
            urllib.parse.quote(token),
            urllib.parse.quote(clientSecret),
            session_id,
        )
        if nextLink:
            # manipulate the nextLink to add the sid, because
            # the caller won't get it until we send a response,
            # by which time we've sent the mail.
            if "?" in nextLink:
                nextLink += "&"
            else:
                nextLink += "?"
            nextLink += "sid=" + urllib.parse.quote(str(session_id))

            link += "&nextLink=%s" % (urllib.parse.quote(nextLink))
        return link

    def validateSessionWithToken(
        self, sid: int, clientSecret: str, token: str
    ) -> Dict[str, bool]:
        """
        Validates the session with the given ID.

        :param sid: The ID of the session to validate.
        :param clientSecret: The client secret to validate.
        :param token: The token to validate.

        :return: A dict with a "success" key which is True if the session
            was successfully validated, False otherwise.
        """
        return common.validateSessionWithToken(self.sydent, sid, clientSecret, token)



# Copyright 2014 OpenMarket Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import attr

# how long a user can wait before validating a session after starting it
THREEPID_SESSION_VALIDATION_TIMEOUT_MS = 24 * 60 * 60 * 1000

# how long we keep sessions for after they've been validated
THREEPID_SESSION_VALID_LIFETIME_MS = 24 * 60 * 60 * 1000


@attr.s(frozen=True, slots=True, auto_attribs=True)
class ValidationSession:
    id: int
    medium: str
    address: str
    client_secret: str
    validated: bool
    mtime: int


@attr.s(frozen=True, slots=True, auto_attribs=True)
class TokenInfo:
    token: str
    send_attempt_number: int


class IncorrectClientSecretException(Exception):
    pass


class SessionExpiredException(Exception):
    pass


class InvalidSessionIdException(Exception):
    pass


class IncorrectSessionTokenException(Exception):
    pass


class SessionNotValidatedException(Exception):
    pass


class DestinationRejectedException(Exception):
    pass



# Copyright 2021 The Matrix.org Foundation C.I.C.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from http import HTTPStatus
from json import JSONDecodeError
from unittest.mock import patch

import twisted.internet.error
import twisted.web.client
from parameterized import parameterized
from twisted.trial import unittest

from tests.utils import make_request, make_sydent


class RegisterTestCase(unittest.TestCase):
    """Tests Sydent's register servlet"""

    def setUp(self) -> None:
        # Create a new sydent
        self.sydent = make_sydent()

    def test_sydent_rejects_invalid_hostname(self) -> None:
        """Tests that the /register endpoint rejects an invalid hostname passed as matrix_server_name"""
        self.sydent.run()

        bad_hostname = "example.com#"

        request, channel = make_request(
            self.sydent.reactor,
            self.sydent.clientApiHttpServer.factory,
            "POST",
            "/_matrix/identity/v2/account/register",
            content={"matrix_server_name": bad_hostname, "access_token": "foo"},
        )

        self.assertEqual(channel.code, 400)

    @parameterized.expand(
        [
            (twisted.internet.error.DNSLookupError(),),
            (twisted.internet.error.TimeoutError(),),
            (twisted.internet.error.ConnectionRefusedError(),),
            # Naughty: strictly we're supposed to initialise a ResponseNeverReceived
            # with a list of 1 or more failures.
            (twisted.web.client.ResponseNeverReceived([]),),
        ]
    )
    def test_connection_failure(self, exc: Exception) -> None:
        self.sydent.run()
        with patch(
            "sydent.http.httpclient.FederationHttpClient.get_json", side_effect=exc
        ):
            request, channel = make_request(
                self.sydent.reactor,
                self.sydent.clientApiHttpServer.factory,
                "POST",
                "/_matrix/identity/v2/account/register",
                content={
                    "matrix_server_name": "matrix.alice.com",
                    "access_token": "back_in_wonderland",
                },
            )
        self.assertEqual(channel.code, HTTPStatus.INTERNAL_SERVER_ERROR)
        self.assertEqual(channel.json_body["errcode"], "M_UNKNOWN")
        # Check that we haven't just returned the generic error message in asyncjsonwrap
        self.assertNotEqual(channel.json_body["error"], "Internal Server Error")
        self.assertIn("contact", channel.json_body["error"])

    def test_federation_does_not_return_json(self) -> None:
        self.sydent.run()
        exc = JSONDecodeError("ruh roh", "C'est n'est pas une objet JSON", 0)
        with patch(
            "sydent.http.httpclient.FederationHttpClient.get_json", side_effect=exc
        ):
            request, channel = make_request(
                self.sydent.reactor,
                self.sydent.clientApiHttpServer.factory,
                "POST",
                "/_matrix/identity/v2/account/register",
                content={
                    "matrix_server_name": "matrix.alice.com",
                    "access_token": "back_in_wonderland",
                },
            )
        self.assertEqual(channel.code, HTTPStatus.INTERNAL_SERVER_ERROR)
        self.assertEqual(channel.json_body["errcode"], "M_UNKNOWN")
        # Check that we haven't just returned the generic error message in asyncjsonwrap
        self.assertNotEqual(channel.json_body["error"], "Internal Server Error")
        self.assertIn("JSON", channel.json_body["error"])


class RegisterAllowListTestCase(unittest.TestCase):
    """
    Test registering works with the `homeserver_allow_list` config option specified
    """

    def test_registering_not_allowed_if_homeserver_not_in_allow_list(self) -> None:
        config = {
            "general": {
                "homeserver_allow_list": "friendly.com, example.com",
                "enable_v1_access": "false",
            }
        }
        # Create a new sydent with a homeserver_allow_list specified
        self.sydent = make_sydent(test_config=config)
        self.sydent.run()

        request, channel = make_request(
            self.sydent.reactor,
            self.sydent.clientApiHttpServer.factory,
            "POST",
            "/_matrix/identity/v2/account/register",
            content={"matrix_server_name": "not.example.com", "access_token": "foo"},
        )
        self.assertEqual(channel.code, 403)
        self.assertEqual(channel.json_body["errcode"], "M_UNAUTHORIZED")



# Copyright 2014 OpenMarket Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING, Dict, List, Optional, Tuple

from sydent.replication.peer import RemotePeer

if TYPE_CHECKING:
    from sydent.sydent import Sydent


class PeerStore:
    def __init__(self, sydent: "Sydent") -> None:
        self.sydent = sydent

    def getPeerByName(self, name: str) -> Optional[RemotePeer]:
        """
        Retrieves a remote peer using it's server name.

        :param name: The server name of the peer.

        :return: The retrieved peer.
        """
        cur = self.sydent.db.cursor()
        res = cur.execute(
            "select p.name, p.port, p.lastSentVersion, pk.alg, pk.key from peers p, peer_pubkeys pk "
            "where p.name = ? and pk.peername = p.name and p.active = 1",
            (name,),
        )

        # Type safety: if the query returns no rows, pubkeys will be empty
        # and we'll return None before using serverName. Otherwise, we'll read
        # at least one row and assign serverName a string value, because the
        # `name` column is declared `not null` in the DB.
        serverName: str = None  # type: ignore[assignment]
        port: Optional[int] = None
        lastSentVer: Optional[int] = None
        pubkeys: Dict[str, str] = {}

        row: Tuple[str, Optional[int], Optional[int], str, str]
        for row in res.fetchall():
            serverName = row[0]
            port = row[1]
            lastSentVer = row[2]
            pubkeys[row[3]] = row[4]

        if len(pubkeys) == 0:
            return None

        p = RemotePeer(self.sydent, serverName, port, pubkeys, lastSentVer)

        return p

    def getAllPeers(self) -> List[RemotePeer]:
        """
        Retrieve all of the remote peers from the database.

        :return: A list of the remote peers this server knows about.
        """
        cur = self.sydent.db.cursor()
        res = cur.execute(
            "select p.name, p.port, p.lastSentVersion, pk.alg, pk.key from peers p, peer_pubkeys pk "
            "where pk.peername = p.name and p.active = 1"
        )

        peers = []

        # Safety: we need to convince ourselves that `peername` will be not None
        # when passed to `RemotePeer`.
        #
        # If `res` is empty, then `pubkeys` will start empty and never be written to.
        # So we will never create a `RemotePeer`. That's fine.
        #
        # Otherwise we process at least one row. The first row we process will
        # satisfy `row[0] is not None` because `name` is nonnull in the schema.
        # `pubkeys` will be empty, so we skip the innermost `if` and assign peername
        # to be a string. There are no further assignments of `None` to `peername`;
        # it will be a string whenever we use it.
        peername: str = None  # type: ignore[assignment]
        port = None
        lastSentVer = None
        pubkeys: Dict[str, str] = {}

        row: Tuple[str, Optional[int], Optional[int], str, str]
        for row in res.fetchall():
            if row[0] != peername:
                if len(pubkeys) > 0:
                    p = RemotePeer(self.sydent, peername, port, pubkeys, lastSentVer)
                    peers.append(p)
                    pubkeys = {}
                peername = row[0]
                port = row[1]
                lastSentVer = row[2]
            pubkeys[row[3]] = row[4]

        if len(pubkeys) > 0:
            p = RemotePeer(self.sydent, peername, port, pubkeys, lastSentVer)
            peers.append(p)

        return peers

    def setLastSentVersionAndPokeSucceeded(
        self,
        peerName: str,
        lastSentVersion: Optional[int],
        lastPokeSucceeded: Optional[int],
    ) -> None:
        """
        Sets the ID of the last association sent to a given peer and the time of the
        last successful request sent to that peer.

        :param peerName: The server name of the peer.
        :param lastSentVersion: The ID of the last association sent to that peer.
        :param lastPokeSucceeded: The timestamp in milliseconds of the last successful
            request sent to that peer.
        """
        cur = self.sydent.db.cursor()
        cur.execute(
            "update peers set lastSentVersion = ?, lastPokeSucceededAt = ? "
            "where name = ?",
            (lastSentVersion, lastPokeSucceeded, peerName),
        )
        self.sydent.db.commit()



from twisted.trial import unittest

from tests.utils import make_sydent


class StartupTestCase(unittest.TestCase):
    """Test that sydent started up correctly"""

    def test_start(self):
        sydent = make_sydent()
        sydent.run()

    def test_homeserver_allow_list_refuses_to_start_if_v1_not_disabled(self):
        """
        Test that Sydent throws a runtime error if `homeserver_allow_list` is specified
        but the v1 API has not been disabled
        """
        config = {
            "general": {
                "homeserver_allow_list": "friendly.com, example.com",
                "enable_v1_access": "true",
            }
        }

        with self.assertRaises(RuntimeError):
            make_sydent(test_config=config)



# Copyright 2019 The Matrix.org Foundation C.I.C.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import urllib
from http import HTTPStatus
from json import JSONDecodeError
from typing import TYPE_CHECKING, Dict

from twisted.internet.error import ConnectError, DNSLookupError
from twisted.web.client import ResponseFailed
from twisted.web.server import Request

from sydent.http.httpclient import FederationHttpClient
from sydent.http.servlets import SydentResource, asyncjsonwrap, get_args, send_cors
from sydent.types import JsonDict
from sydent.users.tokens import issueToken
from sydent.util.stringutils import is_valid_matrix_server_name

if TYPE_CHECKING:
    from sydent.sydent import Sydent

logger = logging.getLogger(__name__)


class RegisterServlet(SydentResource):
    isLeaf = True

    def __init__(self, syd: "Sydent") -> None:
        super().__init__()
        self.sydent = syd
        self.client = FederationHttpClient(self.sydent)

    @asyncjsonwrap
    async def render_POST(self, request: Request) -> JsonDict:
        """
        Register with the Identity Server
        """
        send_cors(request)

        args = get_args(request, ("matrix_server_name", "access_token"))

        matrix_server = args["matrix_server_name"].lower()

        if self.sydent.config.general.homeserver_allow_list:
            if matrix_server not in self.sydent.config.general.homeserver_allow_list:
                request.setResponseCode(403)
                return {
                    "errcode": "M_UNAUTHORIZED",
                    "error": "This homeserver is not authorized to access this server.",
                }

        if not is_valid_matrix_server_name(matrix_server):
            request.setResponseCode(400)
            return {
                "errcode": "M_INVALID_PARAM",
                "error": "matrix_server_name must be a valid Matrix server name (IP address or hostname)",
            }

        def federation_request_problem(error: str) -> Dict[str, str]:
            logger.warning(error)
            request.setResponseCode(HTTPStatus.INTERNAL_SERVER_ERROR)
            return {
                "errcode": "M_UNKNOWN",
                "error": error,
            }

        try:
            result = await self.client.get_json(
                "matrix://%s/_matrix/federation/v1/openid/userinfo?access_token=%s"
                % (
                    matrix_server,
                    urllib.parse.quote(args["access_token"]),
                ),
                1024 * 5,
            )
        except (DNSLookupError, ConnectError, ResponseFailed) as e:
            return federation_request_problem(
                f"Unable to contact the Matrix homeserver ({type(e).__name__})"
            )
        except JSONDecodeError:
            return federation_request_problem(
                "The Matrix homeserver returned invalid JSON"
            )

        if "sub" not in result:
            return federation_request_problem(
                "The Matrix homeserver did not include 'sub' in its response",
            )

        user_id = result["sub"]

        if not isinstance(user_id, str):
            return federation_request_problem(
                "The Matrix homeserver returned a malformed reply"
            )

        user_id_components = user_id.split(":", 1)

        # Ensure there's a localpart and domain in the returned user ID.
        if len(user_id_components) != 2:
            return federation_request_problem(
                "The Matrix homeserver returned an invalid MXID"
            )

        user_id_server = user_id_components[1]

        if not is_valid_matrix_server_name(user_id_server):
            return federation_request_problem(
                "The Matrix homeserver returned an invalid MXID"
            )

        if user_id_server != matrix_server:
            return federation_request_problem(
                "The Matrix homeserver returned a MXID belonging to another homeserver"
            )

        tok = issueToken(self.sydent, user_id)

        # XXX: `token` is correct for the spec, but we released with `access_token`
        # for a substantial amount of time. Serve both to make spec-compliant clients
        # happy.
        return {
            "access_token": tok,
            "token": tok,
        }

    def render_OPTIONS(self, request: Request) -> bytes:
        send_cors(request)
        return b""



# Copyright 2014 OpenMarket Ltd
# Copyright 2018 New Vector Ltd
# Copyright 2019 The Matrix.org Foundation C.I.C.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import TYPE_CHECKING

import twisted.internet.ssl
from twisted.web.resource import Resource
from twisted.web.server import Site

from sydent.http.httpcommon import SizeLimitingRequest
from sydent.http.servlets.accountservlet import AccountServlet
from sydent.http.servlets.authenticated_bind_threepid_servlet import (
    AuthenticatedBindThreePidServlet,
)
from sydent.http.servlets.authenticated_unbind_threepid_servlet import (
    AuthenticatedUnbindThreePidServlet,
)
from sydent.http.servlets.blindlysignstuffservlet import BlindlySignStuffServlet
from sydent.http.servlets.bulklookupservlet import BulkLookupServlet
from sydent.http.servlets.cors_servlet import CorsServlet
from sydent.http.servlets.emailservlet import (
    EmailRequestCodeServlet,
    EmailValidateCodeServlet,
)
from sydent.http.servlets.getvalidated3pidservlet import GetValidated3pidServlet
from sydent.http.servlets.hashdetailsservlet import HashDetailsServlet
from sydent.http.servlets.logoutservlet import LogoutServlet
from sydent.http.servlets.lookupservlet import LookupServlet
from sydent.http.servlets.lookupv2servlet import LookupV2Servlet
from sydent.http.servlets.msisdnservlet import (
    MsisdnRequestCodeServlet,
    MsisdnValidateCodeServlet,
)
from sydent.http.servlets.pubkeyservlets import (
    Ed25519Servlet,
    EphemeralPubkeyIsValidServlet,
    PubkeyIsValidServlet,
)
from sydent.http.servlets.registerservlet import RegisterServlet
from sydent.http.servlets.replication import ReplicationPushServlet
from sydent.http.servlets.store_invite_servlet import StoreInviteServlet
from sydent.http.servlets.termsservlet import TermsServlet
from sydent.http.servlets.threepidbindservlet import ThreePidBindServlet
from sydent.http.servlets.threepidunbindservlet import ThreePidUnbindServlet
from sydent.http.servlets.versions import VersionsServlet

if TYPE_CHECKING:
    from sydent.sydent import Sydent

logger = logging.getLogger(__name__)


class ClientApiHttpServer:
    def __init__(self, sydent: "Sydent", lookup_pepper: str) -> None:
        """
        Args:
            lookup_pepper: The pepper used when hashing identifiers.

        """
        self.sydent = sydent

        root = Resource()
        matrix = Resource()
        identity = Resource()
        api = Resource()
        v1 = CorsServlet(sydent)
        v2 = CorsServlet(sydent)

        validate = Resource()
        validate_v2 = Resource()
        email = Resource()
        email_v2 = Resource()
        msisdn = Resource()
        msisdn_v2 = Resource()

        threepid_v1 = Resource()
        threepid_v2 = Resource()
        unbind = ThreePidUnbindServlet(sydent)

        pubkey = Resource()
        ephemeralPubkey = Resource()

        root.putChild(b"_matrix", matrix)
        matrix.putChild(b"identity", identity)
        identity.putChild(b"api", api)
        identity.putChild(b"v2", v2)
        identity.putChild(b"versions", VersionsServlet())

        pubkey.putChild(b"isvalid", PubkeyIsValidServlet(sydent))
        pubkey.putChild(b"ed25519:0", Ed25519Servlet(sydent))
        pubkey.putChild(b"ephemeral", ephemeralPubkey)
        ephemeralPubkey.putChild(b"isvalid", EphemeralPubkeyIsValidServlet(sydent))

        # v1
        if self.sydent.config.general.enable_v1_access:
            api.putChild(b"v1", v1)
            validate.putChild(b"email", email)
            validate.putChild(b"msisdn", msisdn)
            v1.putChild(b"validate", validate)

            v1.putChild(b"lookup", LookupServlet(sydent))
            v1.putChild(b"bulk_lookup", BulkLookupServlet(sydent))

            v1.putChild(b"pubkey", pubkey)

            threepid_v1.putChild(b"getValidated3pid", GetValidated3pidServlet(sydent))
            threepid_v1.putChild(b"unbind", unbind)
            v1.putChild(b"3pid", threepid_v1)

            email.putChild(b"requestToken", EmailRequestCodeServlet(sydent))
            email.putChild(b"submitToken", EmailValidateCodeServlet(sydent))

            msisdn.putChild(b"requestToken", MsisdnRequestCodeServlet(sydent))
            msisdn.putChild(b"submitToken", MsisdnValidateCodeServlet(sydent))

            v1.putChild(b"store-invite", StoreInviteServlet(sydent))

            v1.putChild(b"sign-ed25519", BlindlySignStuffServlet(sydent))

        if self.sydent.config.general.enable_v1_associations:
            threepid_v1.putChild(b"bind", ThreePidBindServlet(sydent))

        # v2
        # note v2 loses the /api so goes on 'identity' not 'api'
        identity.putChild(b"v2", v2)

        validate_v2.putChild(b"email", email_v2)
        validate_v2.putChild(b"msisdn", msisdn_v2)

        threepid_v2.putChild(
            b"getValidated3pid", GetValidated3pidServlet(sydent, require_auth=True)
        )
        threepid_v2.putChild(b"bind", ThreePidBindServlet(sydent, require_auth=True))
        threepid_v2.putChild(b"unbind", unbind)

        email_v2.putChild(
            b"requestToken", EmailRequestCodeServlet(sydent, require_auth=True)
        )
        email_v2.putChild(
            b"submitToken", EmailValidateCodeServlet(sydent, require_auth=True)
        )

        msisdn_v2.putChild(
            b"requestToken", MsisdnRequestCodeServlet(sydent, require_auth=True)
        )
        msisdn_v2.putChild(
            b"submitToken", MsisdnValidateCodeServlet(sydent, require_auth=True)
        )

        # v2 exclusive APIs
        v2.putChild(b"terms", TermsServlet(sydent))
        account = AccountServlet(sydent)
        v2.putChild(b"account", account)
        account.putChild(b"register", RegisterServlet(sydent))
        account.putChild(b"logout", LogoutServlet(sydent))

        # v2 versions of existing APIs
        v2.putChild(b"validate", validate_v2)
        v2.putChild(b"pubkey", pubkey)
        v2.putChild(b"3pid", threepid_v2)
        v2.putChild(b"store-invite", StoreInviteServlet(sydent, require_auth=True))
        v2.putChild(b"sign-ed25519", BlindlySignStuffServlet(sydent, require_auth=True))
        v2.putChild(b"lookup", LookupV2Servlet(sydent, lookup_pepper))
        v2.putChild(b"hash_details", HashDetailsServlet(sydent, lookup_pepper))

        self.factory = Site(root, SizeLimitingRequest)
        self.factory.displayTracebacks = False

    def setup(self) -> None:
        httpPort = self.sydent.config.http.client_port
        interface = self.sydent.config.http.client_bind_address

        logger.info("Starting Client API HTTP server on %s:%d", interface, httpPort)
        self.sydent.reactor.listenTCP(
            httpPort,
            self.factory,
            backlog=50,  # taken from PosixReactorBase.listenTCP
            interface=interface,
        )


class InternalApiHttpServer:
    def __init__(self, sydent: "Sydent") -> None:
        self.sydent = sydent

    def setup(self, interface: str, port: int) -> None:
        logger.info("Starting Internal API HTTP server on %s:%d", interface, port)
        root = Resource()

        matrix = Resource()
        root.putChild(b"_matrix", matrix)

        identity = Resource()
        matrix.putChild(b"identity", identity)

        internal = Resource()
        identity.putChild(b"internal", internal)

        authenticated_bind = AuthenticatedBindThreePidServlet(self.sydent)
        internal.putChild(b"bind", authenticated_bind)

        authenticated_unbind = AuthenticatedUnbindThreePidServlet(self.sydent)
        internal.putChild(b"unbind", authenticated_unbind)

        factory = Site(root)
        factory.displayTracebacks = False
        self.sydent.reactor.listenTCP(
            port,
            factory,
            backlog=50,  # taken from PosixReactorBase.listenTCP
            interface=interface,
        )


class ReplicationHttpsServer:
    def __init__(self, sydent: "Sydent") -> None:
        self.sydent = sydent

        root = Resource()
        matrix = Resource()
        identity = Resource()

        root.putChild(b"_matrix", matrix)
        matrix.putChild(b"identity", identity)

        replicate = Resource()
        replV1 = Resource()

        identity.putChild(b"replicate", replicate)
        replicate.putChild(b"v1", replV1)
        replV1.putChild(b"push", ReplicationPushServlet(sydent))

        self.factory = Site(root)
        self.factory.displayTracebacks = False

    def setup(self) -> None:
        httpPort = self.sydent.config.http.replication_port
        interface = self.sydent.config.http.replication_bind_address

        if self.sydent.sslComponents.myPrivateCertificate:
            # We will already have logged a warn if this is absent, so don't do it again
            cert = self.sydent.sslComponents.myPrivateCertificate
            certOptions = twisted.internet.ssl.CertificateOptions(
                privateKey=cert.privateKey.original,
                certificate=cert.original,
                trustRoot=self.sydent.sslComponents.trustRoot,
            )

            logger.info("Loaded server private key and certificate!")
            logger.info(
                "Starting Replication HTTPS server on %s:%d", interface, httpPort
            )

            self.sydent.reactor.listenSSL(
                httpPort,
                self.factory,
                certOptions,
                backlog=50,  # taken from PosixReactorBase.listenTCP
                interface=interface,
            )



# Copyright 2016 OpenMarket Ltd
# Copyright 2017 Vector Creations Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import TYPE_CHECKING, Dict, Optional

import phonenumbers

from sydent.db.valsession import ThreePidValSessionStore
from sydent.sms.openmarket import OpenMarketSMS
from sydent.util import time_msec
from sydent.validators import DestinationRejectedException, common

if TYPE_CHECKING:
    from sydent.sydent import Sydent

logger = logging.getLogger(__name__)


class MsisdnValidator:
    def __init__(self, sydent: "Sydent") -> None:
        self.sydent = sydent
        self.omSms = OpenMarketSMS(sydent)

        # cache originators & sms rules from config file
        self.originators = self.sydent.config.sms.originators
        self.smsRules = self.sydent.config.sms.smsRules

    async def requestToken(
        self,
        phoneNumber: phonenumbers.PhoneNumber,
        clientSecret: str,
        send_attempt: int,
        brand: Optional[str] = None,
    ) -> int:
        """
        Creates or retrieves a validation session and sends an text message to the
        corresponding phone number address with a token to use to verify the association.

        :param phoneNumber: The phone number to send the email to.
        :param clientSecret: The client secret to use.
        :param send_attempt: The current send attempt.
        :param brand: A hint at a brand from the request.

        :return: The ID of the session created (or of the existing one if any)
        """
        if str(phoneNumber.country_code) in self.smsRules:
            action = self.smsRules[str(phoneNumber.country_code)]
            if action == "reject":
                raise DestinationRejectedException()

        valSessionStore = ThreePidValSessionStore(self.sydent)

        msisdn = phonenumbers.format_number(
            phoneNumber, phonenumbers.PhoneNumberFormat.E164
        )[1:]

        valSession, token_info = valSessionStore.getOrCreateTokenSession(
            medium="msisdn", address=msisdn, clientSecret=clientSecret
        )

        valSessionStore.setMtime(valSession.id, time_msec())

        if token_info.send_attempt_number >= send_attempt:
            logger.info(
                "Not texting code because current send attempt (%d) is not less than given send attempt (%s)",
                send_attempt,
                token_info.send_attempt_number,
            )
            return valSession.id

        smsBodyTemplate = self.sydent.config.sms.body_template
        originator = self.getOriginator(phoneNumber)

        logger.info(
            "Attempting to text code %s to %s (country %d) with originator %s",
            token_info.token,
            msisdn,
            phoneNumber.country_code,
            originator,
        )

        smsBody = smsBodyTemplate.format(token=token_info.token)

        await self.omSms.sendTextSMS(smsBody, msisdn, originator)

        valSessionStore.setSendAttemptNumber(valSession.id, send_attempt)

        return valSession.id

    def getOriginator(
        self, destPhoneNumber: phonenumbers.PhoneNumber
    ) -> Dict[str, str]:
        """
        Gets an originator for a given phone number.

        :param destPhoneNumber: The phone number to find the originator for.

        :return: The originator (a dict with a "type" key and a "text" key).
        """
        countryCode = str(destPhoneNumber.country_code)

        origs = [
            {
                "type": "alpha",
                "text": "Matrix",
            }
        ]
        if countryCode in self.originators:
            origs = self.originators[countryCode]
        elif "default" in self.originators:
            origs = self.originators["default"]

        # deterministically pick an originator from the list of possible
        # originators, so if someone requests multiple codes, they come from
        # a consistent number (if there's any chance that some originators are
        # more likley to work than others, we may want to change, but it feels
        # like this should be something other than just picking one randomly).
        msisdn = phonenumbers.format_number(
            destPhoneNumber, phonenumbers.PhoneNumberFormat.E164
        )[1:]
        return origs[sum(int(i) for i in msisdn) % len(origs)]

    def validateSessionWithToken(
        self, sid: int, clientSecret: str, token: str
    ) -> Dict[str, bool]:
        """
        Validates the session with the given ID.

        :param sid: The ID of the session to validate.
        :param clientSecret: The client secret to validate.
        :param token: The token to validate.

        :return: A dict with a "success" key which is True if the session
            was successfully validated, False otherwise.
        """
        return common.validateSessionWithToken(self.sydent, sid, clientSecret, token)



# Copyright 2016 OpenMarket Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from base64 import b64encode
from typing import TYPE_CHECKING, Dict, Optional, cast

from twisted.web.http_headers import Headers

from sydent.http.httpclient import SimpleHttpClient
from sydent.sms.types import SendSMSBody, TypeOfNumber
from sydent.types import JsonDict

if TYPE_CHECKING:
    from sydent.sydent import Sydent

logger = logging.getLogger(__name__)


API_BASE_URL = "https://smsc.openmarket.com/sms/v4/mt"
# The Customer Integration Environment, where you can send
# the same requests but it doesn't actually send any SMS.
# Useful for testing.
# API_BASE_URL = "http://smsc-cie.openmarket.com/sms/v4/mt"

# The TON (ie. Type of Number) codes by type used in our config file
TONS: Dict[str, TypeOfNumber] = {
    "long": 1,
    "short": 3,
    "alpha": 5,
}


def tonFromType(t: str) -> TypeOfNumber:
    """
    Get the type of number from the originator's type.

    :param t: Type from the originator.
    :type t: str

    :return: The type of number.
    :rtype: int
    """
    if t in TONS:
        return TONS[t]
    raise Exception("Unknown number type (%s) for originator" % t)


class OpenMarketSMS:
    def __init__(self, sydent: "Sydent") -> None:
        self.sydent = sydent
        self.http_cli = SimpleHttpClient(sydent)

    async def sendTextSMS(
        self, body: str, dest: str, source: Optional[Dict[str, str]] = None
    ) -> None:
        """
        Sends a text message with the given body to the given MSISDN.

        :param body: The message to send.
        :param dest: The destination MSISDN to send the text message to.
        """
        send_body: SendSMSBody = {
            "mobileTerminate": {
                "message": {"content": body, "type": "text"},
                "destination": {
                    "address": dest,
                },
            },
        }
        if source:
            send_body["mobileTerminate"]["source"] = {
                "ton": tonFromType(source["type"]),
                "address": source["text"],
            }

        username = self.sydent.config.sms.api_username
        password = self.sydent.config.sms.api_password

        b64creds = b64encode(b"%s:%s" % (username, password))
        req_headers = Headers(
            {
                b"Authorization": [b"Basic " + b64creds],
                b"Content-Type": [b"application/json"],
            }
        )

        # Type safety: The case from a TypedDict to a regular Dict is required
        # because the two are deliberately not compatible. See
        #    https://github.com/python/mypy/issues/4976
        # for details, but in a nutshell: Dicts can have keys added or removed,
        # and that would break the invariants that a TypedDict is there to check.
        # The case below is safe because we never use send_body afterwards.
        resp, response_body = await self.http_cli.post_json_maybe_get_json(
            API_BASE_URL, cast(JsonDict, send_body), {"headers": req_headers}
        )

        headers = dict(resp.headers.getAllRawHeaders())

        request_id = None
        if b"X-Request-Id" in headers:
            request_id = headers[b"X-Request-Id"][0].decode("UTF-8")

        # Catch errors from the API. The documentation says a success code should be 202
        # Accepted, but let's be broad here just in case and accept all 2xx response
        # codes.
        #
        # Relevant OpenMarket API documentation:
        # https://www.openmarket.com/docs/Content/apis/v4http/send-json.htm
        if resp.code < 200 or resp.code >= 300:
            if response_body is None or "error" not in response_body:
                raise Exception(
                    "OpenMarket API responded with status %d (request ID: %s)"
                    % (
                        resp.code,
                        request_id,
                    ),
                )

            error = response_body["error"]
            raise Exception(
                "OpenMarket API responded with status %d (request ID: %s): %s"
                % (
                    resp.code,
                    request_id,
                    error,
                ),
            )

        ticket_id = None
        if b"Location" not in headers:
            logger.error("Got response from sending SMS with no location header")
        else:
            # Nominally we should parse the URL, but we can just split on '/' since
            # we only care about the last part.
            value = headers[b"Location"][0].decode("UTF-8")
            parts = value.split("/")
            if len(parts) < 2:
                logger.error(
                    "Got response from sending SMS with malformed location header: %s",
                    value,
                )
                return
            else:
                ticket_id = parts[-1]

        logger.info(
            "Successfully sent SMS (ticket ID: %s, request ID %s), OpenMarket API"
            " responded with code %d",
            ticket_id,
            request_id,
            resp.code,
        )



from twisted.trial import unittest

from sydent.util.stringutils import is_valid_matrix_server_name


class UtilTests(unittest.TestCase):
    """Tests Sydent utility functions."""

    def test_is_valid_matrix_server_name(self):
        """Tests that the is_valid_matrix_server_name function accepts only
        valid hostnames (or domain names), with optional port number.
        """
        self.assertTrue(is_valid_matrix_server_name("9.9.9.9"))
        self.assertTrue(is_valid_matrix_server_name("9.9.9.9:4242"))
        self.assertTrue(is_valid_matrix_server_name("[::]"))
        self.assertTrue(is_valid_matrix_server_name("[::]:4242"))
        self.assertTrue(is_valid_matrix_server_name("[a:b:c::]:4242"))

        self.assertTrue(is_valid_matrix_server_name("example.com"))
        self.assertTrue(is_valid_matrix_server_name("EXAMPLE.COM"))
        self.assertTrue(is_valid_matrix_server_name("ExAmPlE.CoM"))
        self.assertTrue(is_valid_matrix_server_name("example.com:4242"))
        self.assertTrue(is_valid_matrix_server_name("localhost"))
        self.assertTrue(is_valid_matrix_server_name("localhost:9000"))
        self.assertTrue(is_valid_matrix_server_name("a.b.c.d:1234"))

        self.assertFalse(is_valid_matrix_server_name("[:::]"))
        self.assertFalse(is_valid_matrix_server_name("a:b:c::"))

        self.assertFalse(is_valid_matrix_server_name("example.com:65536"))
        self.assertFalse(is_valid_matrix_server_name("example.com:0"))
        self.assertFalse(is_valid_matrix_server_name("example.com:-1"))
        self.assertFalse(is_valid_matrix_server_name("example.com:a"))
        self.assertFalse(is_valid_matrix_server_name("example.com: "))
        self.assertFalse(is_valid_matrix_server_name("example.com:04242"))
        self.assertFalse(is_valid_matrix_server_name("example.com: 4242"))
        self.assertFalse(is_valid_matrix_server_name("example.com/example.com"))
        self.assertFalse(is_valid_matrix_server_name("example.com#example.com"))



# Copyright 2018 New Vector Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import time
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple, cast

import attr
import signedjson.key
import signedjson.sign
from signedjson.sign import SignatureVerifyException
from twisted.web.server import Request
from unpaddedbase64 import decode_base64

from sydent.hs_federation.types import (
    CachedVerificationKeys,
    SignedMatrixRequest,
    VerifyKeys,
)
from sydent.http.httpclient import FederationHttpClient
from sydent.types import JsonDict
from sydent.util.stringutils import is_valid_matrix_server_name

if TYPE_CHECKING:
    from sydent.sydent import Sydent

logger = logging.getLogger(__name__)


class NoAuthenticationError(Exception):
    """
    Raised when no signature is provided that could be authenticated
    """

    pass


class InvalidServerName(Exception):
    """
    Raised when the provided origin parameter is not a valid hostname (plus optional port).
    """

    pass


class Verifier:
    """
    Verifies signed json blobs from Matrix Homeservers by finding the
    homeserver's address, contacting it, requesting its keys and
    verifying that the signature on the json blob matches.
    """

    def __init__(self, sydent: "Sydent") -> None:
        self.sydent = sydent
        # Cache of server keys. These are cached until the 'valid_until_ts' time
        # in the result.
        self.cache: Dict[str, CachedVerificationKeys] = {
            # server_name: <result from keys query>,
        }

    async def _getKeysForServer(self, server_name: str) -> VerifyKeys:
        """Get the signing key data from a homeserver.

        :param server_name: The name of the server to request the keys from.

        :return: The verification keys returned by the server.
        """

        if server_name in self.cache:
            cached = self.cache[server_name]
            now = int(time.time() * 1000)
            if cached.valid_until_ts > now:
                return cached.verify_keys

        client = FederationHttpClient(self.sydent)
        # Cast safety: we have validation logic below which checks that
        # - `verify_keys` is present
        # - `valid_until_ts` is an integer if present
        # - cached entries always have a `valid_until_ts` key
        # and we don't use any of the other fields on GetKeyResponse.
        # The only use of the cache is in this function, and only to read
        # the two fields mentioned above.
        result = await client.get_json(
            "matrix://%s/_matrix/key/v2/server/" % server_name, 1024 * 50
        )
        if "verify_keys" not in result:
            raise SignatureVerifyException("No key found in response")

        if not isinstance(result["verify_keys"], dict):
            raise SignatureVerifyException(
                f"Invalid type for verify_keys: expected dict, got {result['verify_keys']}"
            )

        keys_to_remove = []
        for key_name, key_dict in result["verify_keys"].items():
            if "key" not in key_dict:
                logger.warning("Ignoring key %s with no 'key'", key_name)
                keys_to_remove.append(key_name)
            elif not isinstance(key_dict["key"], str):
                raise SignatureVerifyException(
                    f"Invalid type for verify_keys/{key_name}/key: "
                    f"expected str, got {key_dict['key']}"
                )
        for key_name in keys_to_remove:
            del result["verify_keys"][key_name]

        # We've now verified that verify_keys has the correct type.
        verify_keys: VerifyKeys = result["verify_keys"]

        if "valid_until_ts" in result:
            if not isinstance(result["valid_until_ts"], int):
                raise SignatureVerifyException(
                    "Invalid valid_until_ts received, must be an integer"
                )

            # Don't cache anything without a valid_until_ts or we wouldn't
            # know when to expire it.

            logger.info(
                "Got keys for %s: caching until %d",
                server_name,
                result["valid_until_ts"],
            )
            self.cache[server_name] = CachedVerificationKeys(
                verify_keys, result["valid_until_ts"]
            )

        return verify_keys

    async def verifyServerSignedJson(
        self,
        signed_json: SignedMatrixRequest,
        acceptable_server_names: Optional[List[str]] = None,
    ) -> Tuple[str, str]:
        """Given a signed json object, try to verify any one
        of the signatures on it

        XXX: This contains a fairly noddy version of the home server
        SRV lookup and signature verification. It does no caching (just
        fetches the signature each time and does not contact any other
        servers to do perspective checks).

        :param acceptable_server_names: If provided and not None,
        only signatures from servers in this list will be accepted.

        :return a tuple of the server name and key name that was
        successfully verified.

        :raise SignatureVerifyException: The json cannot be verified.
        """
        for server_name, sigs in signed_json.signatures.items():
            if acceptable_server_names is not None:
                if server_name not in acceptable_server_names:
                    continue

            server_keys = await self._getKeysForServer(server_name)
            for key_name, sig in sigs.items():
                if key_name in server_keys:

                    key_bytes = decode_base64(server_keys[key_name]["key"])
                    verify_key = signedjson.key.decode_verify_key_bytes(
                        key_name, key_bytes
                    )
                    logger.info("verifying sig from key %r", key_name)
                    payload = attr.asdict(signed_json)
                    signedjson.sign.verify_signed_json(payload, server_name, verify_key)
                    logger.info(
                        "Verified signature with key %s from %s", key_name, server_name
                    )
                    return (server_name, key_name)
            logger.warning(
                "No matching key found for signature block %r in server keys %r",
                signed_json.signatures,
                server_keys,
            )
        logger.warning(
            "Unable to verify any signatures from block %r. Acceptable server names: %r",
            signed_json.signatures,
            acceptable_server_names,
        )
        raise SignatureVerifyException("No matching signature found")

    async def authenticate_request(self, request: "Request", content: JsonDict) -> str:
        """Authenticates a Matrix federation request based on the X-Matrix header
        XXX: Copied largely from synapse

        :param request: The request object to authenticate
        :param content: The content of the request, if any

        :return: The origin of the server whose signature was validated
        """
        auth_headers = request.requestHeaders.getRawHeaders("Authorization")
        if not auth_headers:
            raise NoAuthenticationError("Missing Authorization headers")

        # Retrieve an origin and signatures from the authorization header.
        origin = None
        signatures: Dict[str, Dict[str, str]] = {}
        for auth in auth_headers:
            if auth.startswith("X-Matrix"):
                (origin, key, sig) = parse_auth_header(auth)
                signatures.setdefault(origin, {})[key] = sig

        if origin is None:
            raise NoAuthenticationError("Missing X-Matrix Authorization header")
        if not is_valid_matrix_server_name(origin):
            raise InvalidServerName(
                "X-Matrix header's origin parameter must be a valid Matrix server name"
            )

        json_request = SignedMatrixRequest(
            method=request.method,
            uri=request.uri,
            destination_is=self.sydent.config.general.server_name,
            signatures=signatures,
            origin=origin,
            content=content,
        )
        await self.verifyServerSignedJson(json_request, [origin])

        logger.info("Verified request from HS %s", origin)

        return origin


def parse_auth_header(header_str: str) -> Tuple[str, str, str]:
    """
    Extracts a server name, signing key and payload signature from an
    "Authorization: X-Matrix ..." header.

    :param header_str: The content of the header, Starting at "X-Matrix".
        For example, `X-Matrix origin=origin.example.com,key="ed25519:key1",sig="ABCDEF..."`
        See https://matrix.org/docs/spec/server_server/r0.1.4#request-authentication

    :return: The server name, the signing key, and the payload signature.

    :raises SignatureVerifyException: if the header did not meet the expected format.
    """
    try:
        # Strip off "X-Matrix " and break up into key-value pairs.
        params = header_str.split(" ")[1].split(",")
        param_dict: Dict[str, str] = dict(
            # Cast safety: the split() call will either return a 1- or 2- tuple.
            # If it returns a 1-tuple, dict() will complain with a ValueError
            # so we'll spot the bad header.
            cast(Tuple[str, str], kv.split("=", maxsplit=1))
            for kv in params
        )

        def strip_quotes(value: str) -> str:
            if value.startswith('"'):
                return value[1:-1]
            else:
                return value

        origin = strip_quotes(param_dict["origin"])
        key = strip_quotes(param_dict["key"])
        sig = strip_quotes(param_dict["sig"])
        return origin, key, sig
    except Exception:
        raise SignatureVerifyException("Malformed Authorization header")



# Copyright 2019 The Matrix.org Foundation C.I.C.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil
import tempfile
import time
from subprocess import Popen

CFG_TEMPLATE = """
[http]
clientapi.http.bind_address = localhost
clientapi.http.port = {port}
client_http_base = http://localhost:{port}
federation.verifycerts = False

[db]
db.file = :memory:

[general]
server.name = test.local
terms.path = {terms_path}
templates.path = {testsubject_path}/res
brand.default = is-test


ip.whitelist = 127.0.0.1

[email]
email.tlsmode = 0
email.invite.subject = %(sender_display_name)s has invited you to chat
email.invite.subject_space = %(sender_display_name)s has invited you to a space
email.smtphost = localhost
email.from = Sydent Validation <noreply@localhost>
email.smtpport = 9925
email.subject = Your Validation Token
email.ratelimit_sender.burst = 100000
email.ratelimit_sender.rate_hz = 100000
"""


class MatrixIsTestLauncher:
    def __init__(self, with_terms):
        self.with_terms = with_terms

    def launch(self):
        sydent_path = os.path.abspath(
            os.path.join(
                os.path.dirname(__file__),
                "..",
            )
        )
        testsubject_path = os.path.join(
            sydent_path,
            "matrix_is_test",
        )
        terms_path = (
            os.path.join(testsubject_path, "terms.yaml") if self.with_terms else ""
        )
        port = 8099 if self.with_terms else 8098

        self.tmpdir = tempfile.mkdtemp(prefix="sydenttest")

        with open(os.path.join(self.tmpdir, "sydent.conf"), "w") as cfgfp:
            cfgfp.write(
                CFG_TEMPLATE.format(
                    testsubject_path=testsubject_path,
                    terms_path=terms_path,
                    port=port,
                )
            )

        newEnv = os.environ.copy()
        newEnv.update(
            {
                "PYTHONPATH": sydent_path,
            }
        )

        stderr_fp = open(os.path.join(testsubject_path, "sydent.stderr"), "w")

        pybin = os.getenv("SYDENT_PYTHON", "python")

        self.process = Popen(
            args=[pybin, "-m", "sydent.sydent"],
            cwd=self.tmpdir,
            env=newEnv,
            stderr=stderr_fp,
        )
        # XXX: wait for startup in a sensible way
        time.sleep(2)

        self._baseUrl = "http://localhost:%d" % (port,)

    def tearDown(self):
        print("Stopping sydent...")
        self.process.terminate()
        shutil.rmtree(self.tmpdir)

    def get_base_url(self):
        return self._baseUrl



# Copyright 2014 OpenMarket Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import time
from typing import NoReturn


def time_msec() -> int:
    """
    Get the current time in milliseconds.

    :return: The current time in milliseconds.
    """
    return int(time.time() * 1000)


def _reject_invalid_json(val: str) -> NoReturn:
    """Do not allow Infinity, -Infinity, or NaN values in JSON."""
    raise ValueError("Invalid JSON value: '%s'" % val)


# a custom JSON decoder which will reject Python extensions to JSON.
json_decoder = json.JSONDecoder(parse_constant=_reject_invalid_json)



#!/usr/bin/env python
# Copyright 2021 The Matrix.org Foundation C.I.C.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import json
import logging
import os
import sqlite3
import sys
import time
from typing import Any, Dict, List, Optional, Tuple

import attr
import signedjson.sign

from sydent.config import SydentConfig
from sydent.sydent import Sydent
from sydent.util import json_decoder
from sydent.util.emailutils import EmailSendException, sendEmail
from sydent.util.hash import sha256_and_url_safe_base64
from tests.utils import ResolvingMemoryReactorClock

logger = logging.getLogger("casefold_db")

# Maximum number of attempts to send an email.
MAX_ATTEMPTS_FOR_EMAIL = 5


@attr.s(auto_attribs=True)
class UpdateDelta:
    """A row to update in the local_threepid_associations table."""

    address: str
    mxid: str
    lookup_hash: str


@attr.s(auto_attribs=True)
class DeleteDelta:
    """A row to delete from the local_threepid_associations table."""

    address: str
    mxid: str


@attr.s(auto_attribs=True)
class Delta:
    """Delta to apply to the local_threepid_associations table for a single
    case-insensitive email address.
    """

    to_update: UpdateDelta
    to_delete: Optional[List[DeleteDelta]] = None


class CantSendEmailException(Exception):
    """Raised when we didn't succeed to send an email after MAX_ATTEMPTS_FOR_EMAIL
    attempts.
    """

    pass


def calculate_lookup_hash(sydent: Sydent, address: str) -> str:
    pepper = sydent.threepidBinder.hashing_store.get_lookup_pepper()
    if pepper is None:
        raise RuntimeError(
            "No lookup pepper found; Sydent should have generated one on startup."
        )
    combo = "%s %s %s" % (address, "email", pepper)
    lookup_hash = sha256_and_url_safe_base64(combo)
    return lookup_hash


def sendEmailWithBackoff(
    sydent: Sydent,
    address: str,
    mxid: str,
    test: bool = False,
) -> None:
    """Send an email with exponential backoff - that way we don't stop sending halfway
    through if the SMTP server rejects our email (e.g. because of rate limiting).

    Setting test to True disables the backoff.

    Raises a CantSendEmailException if no email could be sent after MAX_ATTEMPTS_FOR_EMAIL
    attempts.
    """

    # Disable backoff if we're running tests.
    backoff = 1 if not test else 0

    for i in range(MAX_ATTEMPTS_FOR_EMAIL):
        try:
            template_file = sydent.get_branded_template(
                None,
                "migration_template.eml",
            )

            sendEmail(
                sydent,
                template_file,
                address,
                {"mxid": mxid},
                log_send_errors=False,
            )
            logger.info("Sent email to %s" % address)

            return
        except EmailSendException:
            logger.info(
                "Failed to send email to %s (attempt %d/%d)"
                % (address, i + 1, MAX_ATTEMPTS_FOR_EMAIL)
            )

            time.sleep(backoff)
            backoff *= 2

    raise CantSendEmailException()


def update_local_associations(
    sydent: Sydent,
    db: sqlite3.Connection,
    send_email: bool,
    dry_run: bool,
    test: bool = False,
) -> None:
    """Update the DB table local_threepid_associations so that all stored
    emails are casefolded, and any duplicate mxid's associated with the
    given email are deleted.

    Setting dry_run to True means that the script is being run in dry-run mode
    by the user, i.e. it will run but will not send any email nor update the database.
    Setting test to True means that the function is being called as part of an automated
    test, and therefore we shouldn't backoff when sending emails.

    :return: None
    """

    logger.info("Processing rows in local_threepid_associations")

    res = db.execute(
        "SELECT address, mxid FROM local_threepid_associations WHERE medium = 'email'"
        "ORDER BY ts DESC"
    )

    # a dict that associates an email address with correspoinding mxids and lookup hashes
    associations: Dict[str, List[Tuple[str, str, str]]] = {}

    logger.info("Computing new hashes and signatures for local_threepid_associations")

    # iterate through selected associations, casefold email, rehash it, and add to
    # associations dict
    for address, mxid in res.fetchall():
        casefold_address = address.casefold()

        # rehash email since hashes are case-sensitive
        lookup_hash = calculate_lookup_hash(sydent, casefold_address)

        if casefold_address in associations:
            associations[casefold_address].append((address, mxid, lookup_hash))
        else:
            associations[casefold_address] = [(address, mxid, lookup_hash)]

    # Deltas to apply to the database, associated with the casefolded address they're for.
    deltas: Dict[str, Delta] = {}

    # Iterate through the results, to build the deltas.
    for casefold_address, assoc_tuples in associations.items():
        # If the row is already in the right state and there's no duplicate, don't compute
        # a delta for it.
        if len(assoc_tuples) == 1 and assoc_tuples[0][0] == casefold_address:
            continue

        deltas[casefold_address] = Delta(
            to_update=UpdateDelta(
                address=assoc_tuples[0][0],
                mxid=assoc_tuples[0][1],
                lookup_hash=assoc_tuples[0][2],
            )
        )

        if len(assoc_tuples) > 1:
            # Iterate over all associations except for the first one, since we've already
            # processed it.
            deltas[casefold_address].to_delete = []
            for address, mxid, _ in assoc_tuples[1:]:
                deltas[casefold_address].to_delete.append(
                    DeleteDelta(
                        address=address,
                        mxid=mxid,
                    )
                )

    logger.info(f"{len(deltas)} rows to update in local_threepid_associations")

    # Apply the deltas
    for casefolded_address, delta in deltas.items():
        if not test:
            log_msg = f"Updating {casefolded_address}"
            if delta.to_delete is not None:
                log_msg += (
                    f" and deleting {len(delta.to_delete)} rows associated with it"
                )
            logger.info(log_msg)

        try:
            # Delete each association, and send an email mentioning the affected MXID.
            if delta.to_delete is not None and not dry_run:
                for to_delete in delta.to_delete:
                    if send_email and to_delete.mxid != delta.to_update.mxid:
                        # If the MXID is one that will still be associated with this
                        # email address after this run, don't send an email for it.
                        sendEmailWithBackoff(
                            sydent,
                            to_delete.address,
                            to_delete.mxid,
                            test=test,
                        )

                    logger.debug(
                        "Deleting %s from table local_threepid_associations",
                        to_delete.address,
                    )
                    cur = db.cursor()
                    cur.execute(
                        "DELETE FROM local_threepid_associations WHERE medium = 'email' AND address = ?",
                        (to_delete.address,),
                    )
                    db.commit()

            # Update the row now that there's no duplicate.
            if not dry_run:
                logger.debug(
                    "Updating table local threepid associations setting address to %s, "
                    "lookup_hash to %s, where medium = email and address = %s and mxid = %s",
                    casefolded_address,
                    delta.to_update.lookup_hash,
                    delta.to_update.address,
                    delta.to_update.mxid,
                )
                cur = db.cursor()
                cur.execute(
                    "UPDATE local_threepid_associations SET address = ?, lookup_hash = ? WHERE medium = 'email' AND address = ? AND mxid = ?",
                    (
                        casefolded_address,
                        delta.to_update.lookup_hash,
                        delta.to_update.address,
                        delta.to_update.mxid,
                    ),
                )
                db.commit()

        except CantSendEmailException:
            # If we failed because we couldn't send an email move on to the next address
            # to de-duplicate.
            # We catch this error here rather than when sending the email because we want
            # to avoid deleting rows we can't warn users about, and we don't want to
            # proceed with the subsequent update because there might still be duplicates
            # in the database (since we haven't deleted everything we wanted to delete).
            logger.warn("Failed to send email to %s; skipping!", to_delete.address)
            continue


def update_global_associations(
    sydent: Sydent,
    db: sqlite3.Connection,
    dry_run: bool,
) -> None:
    """Update the DB table global_threepid_associations so that all stored
    emails are casefolded, the signed association is re-signed and any duplicate
    mxid's associated with the given email are deleted.

    Setting dry_run to True means that the script is being run in dry-run mode
    by the user, i.e. it will run but will not send any email nor update the database.

    :return: None
    """

    logger.info("Processing rows in global_threepid_associations")

    # get every row where the local server is origin server and medium is email
    origin_server = sydent.config.general.server_name
    medium = "email"

    res = db.execute(
        "SELECT address, mxid, sgAssoc FROM global_threepid_associations WHERE medium = ?"
        "AND originServer = ? ORDER BY ts DESC",
        (medium, origin_server),
    )

    # dict that stores email address with mxid, email address, lookup hash, and
    # signed association
    associations: Dict[str, List[Tuple[str, str, str, str]]] = {}

    logger.info("Computing new hashes and signatures for global_threepid_associations")

    # iterate through selected associations, casefold email, rehash it, re-sign the
    # associations and add to associations dict
    for address, mxid, sg_assoc in res.fetchall():
        casefold_address = address.casefold()

        # rehash the email since hash functions are case-sensitive
        lookup_hash = calculate_lookup_hash(sydent, casefold_address)

        # update signed associations with new casefolded address and re-sign
        sg_assoc = json_decoder.decode(sg_assoc)
        sg_assoc["address"] = address.casefold()
        sg_assoc = json.dumps(
            signedjson.sign.sign_json(
                sg_assoc, sydent.config.general.server_name, sydent.keyring.ed25519
            )
        )

        if casefold_address in associations:
            associations[casefold_address].append(
                (address, mxid, lookup_hash, sg_assoc)
            )
        else:
            associations[casefold_address] = [(address, mxid, lookup_hash, sg_assoc)]

    # list of arguments to update db with
    db_update_args: List[Tuple[Any, str, str, str, str]] = []

    # list of mxids to delete
    to_delete: List[Tuple[str]] = []

    for casefold_address, assoc_tuples in associations.items():
        # If the row is already in the right state and there's no duplicate, don't compute
        # a delta for it.
        if len(assoc_tuples) == 1 and assoc_tuples[0][0] == casefold_address:
            continue

        db_update_args.append(
            (
                casefold_address,
                assoc_tuples[0][2],
                assoc_tuples[0][3],
                assoc_tuples[0][0],
                assoc_tuples[0][1],
            )
        )

        if len(assoc_tuples) > 1:
            # Iterate over all associations except for the first one, since we've already
            # processed it.
            for address, mxid, _, _ in assoc_tuples[1:]:
                to_delete.append((address,))

    logger.info(
        f"{len(to_delete)} rows to delete, {len(db_update_args)} rows to update in global_threepid_associations"
    )
    if not dry_run:
        cur = db.cursor()

        if len(to_delete) > 0:
            cur.executemany(
                "DELETE FROM global_threepid_associations WHERE medium = 'email' AND address = ?",
                to_delete,
            )

            logger.info(
                f"{len(to_delete)} rows deleted from global_threepid_associations"
            )

        if len(db_update_args) > 0:
            cur.executemany(
                "UPDATE global_threepid_associations SET address = ?, lookup_hash = ?, sgAssoc = ? WHERE medium = 'email' AND address = ? AND mxid = ?",
                db_update_args,
            )

            logger.info(
                f"{len(db_update_args)} rows updated in global_threepid_associations"
            )

        db.commit()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Casefold email addresses in database")
    parser.add_argument(
        "--no-email", action="store_true", help="run script but do not send emails"
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="run script but do not send emails or alter database",
    )

    parser.add_argument("config_path", help="path to the sydent configuration file")

    args = parser.parse_args()

    # Set up logging.
    log_format = "%(asctime)s - %(name)s - %(lineno)d - %(levelname)s" " - %(message)s"
    formatter = logging.Formatter(log_format)
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    logger.setLevel(logging.INFO)
    logger.addHandler(handler)

    # if the path the user gives us doesn't work, find it for them
    if not os.path.exists(args.config_path):
        logger.error(f"The config file '{args.config_path}' does not exist.")
        sys.exit(1)

    sydent_config = SydentConfig()
    sydent_config.parse_config_file(args.config_path)

    reactor = ResolvingMemoryReactorClock()
    sydent = Sydent(sydent_config, reactor, False)

    update_global_associations(sydent, sydent.db, dry_run=args.dry_run)
    update_local_associations(
        sydent,
        sydent.db,
        send_email=not args.no_email,
        dry_run=args.dry_run,
    )



# Copyright 2014 OpenMarket Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING, Optional

from twisted.web.server import Request

from sydent.http.auth import authV2
from sydent.http.servlets import SydentResource, get_args, jsonwrap, send_cors
from sydent.types import JsonDict
from sydent.util.emailutils import EmailAddressException, EmailSendException
from sydent.util.stringutils import MAX_EMAIL_ADDRESS_LENGTH, is_valid_client_secret
from sydent.validators import (
    IncorrectClientSecretException,
    IncorrectSessionTokenException,
    InvalidSessionIdException,
    SessionExpiredException,
)

if TYPE_CHECKING:
    from sydent.sydent import Sydent


class EmailRequestCodeServlet(SydentResource):
    isLeaf = True

    def __init__(self, syd: "Sydent", require_auth: bool = False) -> None:
        super().__init__()
        self.sydent = syd
        self.require_auth = require_auth

    @jsonwrap
    def render_POST(self, request: Request) -> JsonDict:
        send_cors(request)

        ipaddress = self.sydent.ip_from_request(request)

        if self.require_auth:
            account = authV2(self.sydent, request)

            self.sydent.email_sender_ratelimiter.ratelimit(account.userId)
        elif ipaddress:
            # For `/v1/` requests the ip address is the best we can do for rate
            # limiting.
            self.sydent.email_sender_ratelimiter.ratelimit(ipaddress)

        args = get_args(request, ("email", "client_secret", "send_attempt"))

        email = args["email"]
        clientSecret = args["client_secret"]

        try:
            # if we got this via the v1 API in a querystring or urlencoded body,
            # then the values in args will be a string. So check that
            # send_attempt is an int.
            #
            # NB: We don't check if we're processing a url-encoded v1 request.
            # This means we accept string representations of integers for
            # `send_attempt` in v2 requests, and in v1 requests that supply a
            # JSON body. This is contrary to the spec and leaves me with a dirty
            # feeling I can't quite shake off.
            #
            # Where's Raymond Hettinger when you need him? (THUMP) There must be
            # a better way!
            sendAttempt = int(args["send_attempt"])
        except (TypeError, ValueError):
            request.setResponseCode(400)
            return {
                "errcode": "M_INVALID_PARAM",
                "error": f"send_attempt should be an integer (got {args['send_attempt']}",
            }

        if not is_valid_client_secret(clientSecret):
            request.setResponseCode(400)
            return {
                "errcode": "M_INVALID_PARAM",
                "error": "Invalid client_secret provided",
            }

        if not (0 < len(email) <= MAX_EMAIL_ADDRESS_LENGTH):
            request.setResponseCode(400)
            return {"errcode": "M_INVALID_PARAM", "error": "Invalid email provided"}

        brand = self.sydent.brand_from_request(request)

        nextLink: Optional[str] = None
        if "next_link" in args and not args["next_link"].startswith("file:///"):
            nextLink = args["next_link"]

        try:
            sid = self.sydent.validators.email.requestToken(
                email,
                clientSecret,
                sendAttempt,
                nextLink,
                ipaddress=ipaddress,
                brand=brand,
            )
            resp = {"sid": str(sid)}
        except EmailAddressException:
            request.setResponseCode(400)
            resp = {"errcode": "M_INVALID_EMAIL", "error": "Invalid email address"}
        except EmailSendException:
            request.setResponseCode(500)
            resp = {"errcode": "M_EMAIL_SEND_ERROR", "error": "Failed to send email"}

        return resp

    def render_OPTIONS(self, request: Request) -> bytes:
        send_cors(request)
        return b""


class EmailValidateCodeServlet(SydentResource):
    isLeaf = True

    def __init__(self, syd: "Sydent", require_auth: bool = False) -> None:
        super().__init__()
        self.sydent = syd
        self.require_auth = require_auth

    def render_GET(self, request: Request) -> bytes:
        args = get_args(request, ("nextLink",), required=False)

        resp = None
        try:
            resp = self.do_validate_request(request)
        except Exception:
            pass
        if resp and "success" in resp and resp["success"]:
            msg = "Verification successful! Please return to your Matrix client to continue."
            if "nextLink" in args:
                next_link = args["nextLink"]
                if not next_link.startswith("file:///"):
                    request.setResponseCode(302)
                    request.setHeader("Location", next_link)
        else:
            msg = "Verification failed: you may need to request another verification email"

        brand = self.sydent.brand_from_request(request)

        # self.sydent.config.http.verify_response_template is deprecated
        if self.sydent.config.http.verify_response_template is None:
            templateFile = self.sydent.get_branded_template(
                brand,
                "verify_response_template.html",
            )
        else:
            templateFile = self.sydent.config.http.verify_response_template

        request.setHeader("Content-Type", "text/html")
        res = open(templateFile).read() % {"message": msg}

        return res.encode("UTF-8")

    @jsonwrap
    def render_POST(self, request: Request) -> JsonDict:
        send_cors(request)

        if self.require_auth:
            authV2(self.sydent, request)

        return self.do_validate_request(request)

    def do_validate_request(self, request: Request) -> JsonDict:
        """
        Extracts information about a validation session from the request and
        attempts to validate that session.

        :param request: The request to extract information about the session from.

        :return: A dict with a "success" key which value indicates whether the
            validation succeeded. If the validation failed, this dict also includes
            a "errcode" and a "error" keys which include information about the failure.
        """
        args = get_args(request, ("token", "sid", "client_secret"))

        sid = args["sid"]
        tokenString = args["token"]
        clientSecret = args["client_secret"]

        if not is_valid_client_secret(clientSecret):
            request.setResponseCode(400)
            return {
                "errcode": "M_INVALID_PARAM",
                "error": "Invalid client_secret provided",
            }

        try:
            return self.sydent.validators.email.validateSessionWithToken(
                sid, clientSecret, tokenString
            )
        except IncorrectClientSecretException:
            return {
                "success": False,
                "errcode": "M_INVALID_PARAM",
                "error": "Client secret does not match the one given when requesting the token",
            }
        except SessionExpiredException:
            return {
                "success": False,
                "errcode": "M_SESSION_EXPIRED",
                "error": "This validation session has expired: call requestToken again",
            }
        except InvalidSessionIdException:
            return {
                "success": False,
                "errcode": "M_INVALID_PARAM",
                "error": "The token doesn't match",
            }
        except IncorrectSessionTokenException:
            return {
                "success": False,
                "errcode": "M_NO_VALID_SESSION",
                "error": "No session could be found with this sid",
            }

    def render_OPTIONS(self, request: Request) -> bytes:
        send_cors(request)
        return b""



# Copyright 2021 Matrix.org Foundation C.I.C.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from unittest.mock import patch

from parameterized import parameterized
from twisted.trial import unittest

from sydent.users.accounts import Account
from tests.utils import make_request, make_sydent


class StoreInviteTestCase(unittest.TestCase):
    """Tests Sydent's register servlet"""

    def setUp(self) -> None:
        # Create a new sydent
        config = {
            "email": {
                "email.from": "Sydent Validation <noreply@hostname>",
            },
        }
        self.sydent = make_sydent(test_config=config)
        self.sender = "@alice:wonderland"

    @parameterized.expand(
        [
            ("not@an@email@address",),
            ("Naughty Nigel <perfectly.valid@mail.address>",),
        ]
    )
    def test_invalid_email_returns_400(self, address: str) -> None:
        self.sydent.run()

        with patch("sydent.http.servlets.store_invite_servlet.authV2") as authV2:
            authV2.return_value = Account(self.sender, 0, None)

            request, channel = make_request(
                self.sydent.reactor,
                self.sydent.clientApiHttpServer.factory,
                "POST",
                "/_matrix/identity/v2/store-invite",
                content={
                    "address": address,
                    "medium": "email",
                    "room_id": "!myroom:test",
                    "sender": self.sender,
                },
            )

        self.assertEqual(channel.code, 400)
        self.assertEqual(channel.json_body["errcode"], "M_INVALID_EMAIL")



# Copyright 2014-2015 OpenMarket Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import email.utils
import logging
import random
import smtplib
import ssl
import string
import urllib
from html import escape
from typing import TYPE_CHECKING, Dict

import twisted.python.log
from prometheus_client import Counter

from sydent.util import time_msec
from sydent.util.tokenutils import generateAlphanumericTokenOfLength

if TYPE_CHECKING:
    from sydent.sydent import Sydent

logger = logging.getLogger(__name__)

email_counter = Counter("sydent_emails_sent", "Number of emails we attempted to send")


def sendEmail(
    sydent: "Sydent",
    templateFile: str,
    mailTo: str,
    substitutions: Dict[str, str],
    log_send_errors: bool = True,
) -> None:
    """
    Sends an email with the given parameters.

    :param sydent: The Sydent instance to use when building the configuration to send the
        email with.
    :param templateFile: The filename of the template to use when building the body of the
        email.
    :param mailTo: The email address to send the email to.
    :param substitutions: The substitutions to use with the template.
    :param log_send_errors: Whether to log errors happening when sending an email.
    """
    mailFrom = sydent.config.email.sender
    myHostname = sydent.config.email.host_name

    midRandom = "".join([random.choice(string.ascii_letters) for _ in range(16)])
    messageid = "<%d%s@%s>" % (time_msec(), midRandom, myHostname)

    substitutions.update(
        {
            "messageid": messageid,
            "date": email.utils.formatdate(localtime=False),
            "to": mailTo,
            "from": mailFrom,
        }
    )

    # use jinja for rendering if jinja templates are present
    if templateFile.endswith(".j2"):
        # We add randomize the multipart boundary to stop user input from
        # conflicting with it.
        substitutions["multipart_boundary"] = generateAlphanumericTokenOfLength(32)
        template = sydent.config.general.template_environment.get_template(templateFile)
        mailString = template.render(substitutions)
    else:
        allSubstitutions = {}
        for k, v in substitutions.items():
            allSubstitutions[k] = v
            allSubstitutions[k + "_forhtml"] = escape(v)
            allSubstitutions[k + "_forurl"] = urllib.parse.quote(v)
        allSubstitutions["multipart_boundary"] = generateAlphanumericTokenOfLength(32)
        with open(templateFile) as template_file:
            mailString = template_file.read() % allSubstitutions

    try:
        check_valid_email_address(mailTo, allow_description=False)
    except EmailAddressException:
        logger.warning("Invalid email address %s", mailTo)
        raise

    mailServer = sydent.config.email.smtp_server
    mailPort = int(sydent.config.email.smtp_port)
    mailUsername = sydent.config.email.smtp_username
    mailPassword = sydent.config.email.smtp_password
    mailTLSMode = sydent.config.email.tls_mode

    logger.info(
        "Sending mail to %s with mail server: %s"
        % (
            mailTo,
            mailServer,
        )
    )
    try:
        smtp: smtplib.SMTP
        # Explicitly create a context, to ensure we verify the server's certificate
        # and hostname.
        ctx = ssl.create_default_context(purpose=ssl.Purpose.SERVER_AUTH)
        if mailTLSMode == "SSL" or mailTLSMode == "TLS":
            smtp = smtplib.SMTP_SSL(mailServer, mailPort, myHostname, context=ctx)
        elif mailTLSMode == "STARTTLS":
            smtp = smtplib.SMTP(mailServer, mailPort, myHostname)
            smtp.starttls(context=ctx)
        else:
            smtp = smtplib.SMTP(mailServer, mailPort, myHostname)
        if mailUsername != "":
            smtp.login(mailUsername, mailPassword)

        email_counter.inc()

        # We're using the parsing above to do basic validation, but instead of
        # failing it may munge the address it returns. So we should *not* use
        # that parsed address, as it may not match any validation done
        # elsewhere.
        smtp.sendmail(mailFrom, mailTo, mailString.encode("utf-8"))
        smtp.quit()
    except Exception as origException:
        if log_send_errors:
            twisted.python.log.err()
        raise EmailSendException() from origException


def check_valid_email_address(address: str, allow_description: bool) -> None:
    """Check the given string is a valid email address.

    Email addresses are complicated (see RFCs 5321, 5322 and 6531; plus
    https://www.netmeister.org/blog/email.html). This isn't a comprehensive
    validation; we defer to Python's stdlib.

    :raises EmailAddressException: if not.
    """
    parsed_address = email.utils.parseaddr(address)[1]
    if parsed_address == "":
        raise EmailAddressException(f"Couldn't parse email address {address}.")
    if not allow_description and address != parsed_address:
        raise EmailAddressException(
            f"Parsing address ({address} yielded a different address"
            f"({parsed_address})"
        )


class EmailAddressException(Exception):
    pass


class EmailSendException(Exception):
    pass



# Copyright 2021 The Matrix.org Foundation C.I.C.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from configparser import ConfigParser
from typing import Dict, List

from sydent.config._base import BaseConfig
from sydent.config.exceptions import ConfigError


class SMSConfig(BaseConfig):
    def parse_config(self, cfg: "ConfigParser") -> bool:
        """
        Parse the sms section of the config

        :param cfg: the configuration to be parsed
        """
        self.body_template = cfg.get("sms", "bodyTemplate")

        # Make sure username and password are bytes otherwise we can't use them with
        # b64encode.
        self.api_username = cfg.get("sms", "username").encode("UTF-8")
        self.api_password = cfg.get("sms", "password").encode("UTF-8")

        self.originators: Dict[str, List[Dict[str, str]]] = {}
        self.smsRules = {}

        for opt in cfg.options("sms"):
            if opt.startswith("originators."):
                country = opt.split(".")[1]
                rawVal = cfg.get("sms", opt)
                rawList = [i.strip() for i in rawVal.split(",")]

                self.originators[country] = []
                for origString in rawList:
                    parts = origString.split(":")
                    if len(parts) != 2:
                        raise ConfigError(
                            "Originators must be in form: long:<number>, short:<number> or alpha:<text>, separated by commas"
                        )
                    if parts[0] not in ["long", "short", "alpha"]:
                        raise ConfigError(
                            "Invalid originator type: valid types are long, short and alpha"
                        )
                    self.originators[country].append(
                        {
                            "type": parts[0],
                            "text": parts[1],
                        }
                    )
            elif opt.startswith("smsrule."):
                country = opt.split(".")[1]
                action = cfg.get("sms", opt)

                if action not in ["allow", "reject"]:
                    raise ConfigError(
                        "Invalid SMS rule action: %s, expecting 'allow' or 'reject'"
                        % action
                    )

                self.smsRules[country] = action

        self.msisdn_ratelimit_burst = cfg.getint(
            "sms", "msisdn.ratelimit.burst", fallback=5
        )
        self.msisdn_ratelimit_rate_hz = cfg.getfloat(
            "sms", "msisdn.ratelimit.rate_hz", fallback=1.0 / (60.0 * 60.0)
        )

        self.country_ratelimit_burst = cfg.getint(
            "sms", "country.ratelimit.burst", fallback=50
        )
        self.country_ratelimit_rate_hz = cfg.getfloat(
            "sms", "country.ratelimit.rate_hz", fallback=1.0 / 60.0
        )

        return False



# Copyright 2021 The Matrix.org Foundation C.I.C.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Any, List, Optional

from netaddr import IPAddress, IPSet
from twisted.internet.address import IPv4Address, IPv6Address
from twisted.internet.interfaces import (
    IAddress,
    IHostResolution,
    IReactorPluggableNameResolver,
    IResolutionReceiver,
)
from zope.interface import implementer, provider

logger = logging.getLogger(__name__)


def check_against_blacklist(
    ip_address: IPAddress, ip_whitelist: Optional[IPSet], ip_blacklist: IPSet
) -> bool:
    """
    Compares an IP address to allowed and disallowed IP sets.

    Args:
        ip_address: The IP address to check
        ip_whitelist: Allowed IP addresses.
        ip_blacklist: Disallowed IP addresses.

    Returns:
        True if the IP address is in the blacklist and not in the whitelist.
    """
    if ip_address in ip_blacklist:
        if ip_whitelist is None or ip_address not in ip_whitelist:
            return True
    return False


class _IPBlacklistingResolver:
    """
    A proxy for reactor.nameResolver which only produces non-blacklisted IP
    addresses, preventing DNS rebinding attacks on URL preview.
    """

    def __init__(
        self,
        reactor: IReactorPluggableNameResolver,
        ip_whitelist: Optional[IPSet],
        ip_blacklist: IPSet,
    ):
        """
        Args:
            reactor: The twisted reactor.
            ip_whitelist: IP addresses to allow.
            ip_blacklist: IP addresses to disallow.
        """
        self._reactor = reactor
        self._ip_whitelist = ip_whitelist
        self._ip_blacklist = ip_blacklist

    def resolveHostName(
        self, recv: IResolutionReceiver, hostname: str, portNumber: int = 0
    ) -> IResolutionReceiver:
        addresses = []  # type: List[IAddress]

        def _callback() -> None:
            has_bad_ip = False
            for address in addresses:
                # We only expect IPv4 and IPv6 addresses since only A/AAAA lookups
                # should go through this path.
                if not isinstance(address, (IPv4Address, IPv6Address)):
                    continue

                ip_address = IPAddress(address.host)

                if check_against_blacklist(
                    ip_address, self._ip_whitelist, self._ip_blacklist
                ):
                    logger.info(
                        "Dropped %s from DNS resolution to %s due to blacklist"
                        % (ip_address, hostname)
                    )
                    has_bad_ip = True

            # if we have a blacklisted IP, we'd like to raise an error to block the
            # request, but all we can really do from here is claim that there were no
            # valid results.
            if not has_bad_ip:
                for address in addresses:
                    recv.addressResolved(address)
            recv.resolutionComplete()

        @provider(IResolutionReceiver)
        class EndpointReceiver:
            @staticmethod
            def resolutionBegan(resolutionInProgress: IHostResolution) -> None:
                recv.resolutionBegan(resolutionInProgress)

            @staticmethod
            def addressResolved(address: IAddress) -> None:
                addresses.append(address)

            @staticmethod
            def resolutionComplete() -> None:
                _callback()

        self._reactor.nameResolver.resolveHostName(
            EndpointReceiver, hostname, portNumber=portNumber
        )

        return recv


@implementer(IReactorPluggableNameResolver)
class BlacklistingReactorWrapper:
    """
    A Reactor wrapper which will prevent DNS resolution to blacklisted IP
    addresses, to prevent DNS rebinding.
    """

    def __init__(
        self,
        reactor: IReactorPluggableNameResolver,
        ip_whitelist: Optional[IPSet],
        ip_blacklist: IPSet,
    ):
        self._reactor = reactor

        # We need to use a DNS resolver which filters out blacklisted IP
        # addresses, to prevent DNS rebinding.
        self.nameResolver = _IPBlacklistingResolver(
            self._reactor, ip_whitelist, ip_blacklist
        )

    def __getattr__(self, attr: str) -> Any:
        # Passthrough to the real reactor except for the DNS resolver.
        return getattr(self._reactor, attr)



# Copyright 2019 The Matrix.org Foundation C.I.C.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import TYPE_CHECKING

from twisted.web.server import Request

from sydent.db.accounts import AccountStore
from sydent.http.auth import authV2, tokenFromRequest
from sydent.http.servlets import MatrixRestError, SydentResource, jsonwrap, send_cors
from sydent.types import JsonDict

if TYPE_CHECKING:
    from sydent.sydent import Sydent

logger = logging.getLogger(__name__)


class LogoutServlet(SydentResource):
    isLeaf = True

    def __init__(self, syd: "Sydent") -> None:
        super().__init__()
        self.sydent = syd

    @jsonwrap
    def render_POST(self, request: Request) -> JsonDict:
        """
        Invalidate the given access token
        """
        send_cors(request)

        authV2(self.sydent, request, False)

        token = tokenFromRequest(request)
        if token is None:
            raise MatrixRestError(400, "M_MISSING_PARAMS", "Missing token")

        accountStore = AccountStore(self.sydent)
        accountStore.delToken(token)
        return {}

    def render_OPTIONS(self, request: Request) -> bytes:
        send_cors(request)
        return b""



#  Copyright 2021 The Matrix.org Foundation C.I.C.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import itertools
from typing import Iterable, Optional

from netaddr import AddrFormatError, IPNetwork, IPSet

# IP ranges that are considered private / unroutable / don't make sense.
DEFAULT_IP_RANGE_BLACKLIST = [
    # Localhost
    "127.0.0.0/8",
    # Private networks.
    "10.0.0.0/8",
    "172.16.0.0/12",
    "192.168.0.0/16",
    # Carrier grade NAT.
    "100.64.0.0/10",
    # Address registry.
    "192.0.0.0/24",
    # Link-local networks.
    "169.254.0.0/16",
    # Formerly used for 6to4 relay.
    "192.88.99.0/24",
    # Testing networks.
    "198.18.0.0/15",
    "192.0.2.0/24",
    "198.51.100.0/24",
    "203.0.113.0/24",
    # Multicast.
    "224.0.0.0/4",
    # Localhost
    "::1/128",
    # Link-local addresses.
    "fe80::/10",
    # Unique local addresses.
    "fc00::/7",
    # Testing networks.
    "2001:db8::/32",
    # Multicast.
    "ff00::/8",
    # Site-local addresses
    "fec0::/10",
]


def generate_ip_set(
    ip_addresses: Optional[Iterable[str]],
    extra_addresses: Optional[Iterable[str]] = None,
    config_path: Optional[Iterable[str]] = None,
) -> IPSet:
    """
    Generate an IPSet from a list of IP addresses or CIDRs.

    Additionally, for each IPv4 network in the list of IP addresses, also
    includes the corresponding IPv6 networks.

    This includes:

    * IPv4-Compatible IPv6 Address (see RFC 4291, section 2.5.5.1)
    * IPv4-Mapped IPv6 Address (see RFC 4291, section 2.5.5.2)
    * 6to4 Address (see RFC 3056, section 2)

    Args:
        ip_addresses: An iterable of IP addresses or CIDRs.
        extra_addresses: An iterable of IP addresses or CIDRs.
        config_path: The path in the configuration for error messages.

    Returns:
        A new IP set.
    """
    result = IPSet()
    for ip in itertools.chain(ip_addresses or (), extra_addresses or ()):
        try:
            network = IPNetwork(ip)
        except AddrFormatError as e:
            raise Exception(
                "Invalid IP range provided: %s." % (ip,), config_path
            ) from e
        result.add(network)

        # It is possible that these already exist in the set, but that's OK.
        if ":" not in str(network):
            result.add(IPNetwork(network).ipv6(ipv4_compatible=True))
            result.add(IPNetwork(network).ipv6(ipv4_compatible=False))
            result.add(_6to4(network))

    return result


def _6to4(network: IPNetwork) -> IPNetwork:
    """Convert an IPv4 network into a 6to4 IPv6 network per RFC 3056."""

    # 6to4 networks consist of:
    # * 2002 as the first 16 bits
    # * The first IPv4 address in the network hex-encoded as the next 32 bits
    # * The new prefix length needs to include the bits from the 2002 prefix.
    hex_network = hex(network.first)[2:]
    hex_network = ("0" * (8 - len(hex_network))) + hex_network
    return IPNetwork(
        "2002:%s:%s::/%d"
        % (
            hex_network[:4],
            hex_network[4:],
            16 + network.prefixlen,
        )
    )



# Copyright 2016 OpenMarket Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
from io import BytesIO
from typing import TYPE_CHECKING, Any, Dict, Generic, Optional, Tuple, TypeVar, cast

from twisted.web.client import Agent, FileBodyProducer
from twisted.web.http_headers import Headers
from twisted.web.iweb import IAgent, IResponse

from sydent.http.blacklisting_reactor import BlacklistingReactorWrapper
from sydent.http.federation_tls_options import ClientTLSOptionsFactory
from sydent.http.httpcommon import read_body_with_max_size
from sydent.http.matrixfederationagent import MatrixFederationAgent
from sydent.types import JsonDict
from sydent.util import json_decoder

if TYPE_CHECKING:
    from sydent.sydent import Sydent

logger = logging.getLogger(__name__)


AgentType = TypeVar("AgentType", bound=IAgent)


class HTTPClient(Generic[AgentType]):
    """A base HTTP class that contains methods for making GET and POST HTTP
    requests.
    """

    agent: AgentType

    async def get_json(self, uri: str, max_size: Optional[int] = None) -> JsonDict:
        """Make a GET request to an endpoint returning JSON and parse result

        :param uri: The URI to make a GET request to.

        :param max_size: The maximum size (in bytes) to allow as a response.

        :return: A deferred containing JSON parsed into a Python object.
        """
        logger.debug("HTTP GET %s", uri)

        response = await self.agent.request(
            b"GET",
            uri.encode("utf8"),
        )
        body = await read_body_with_max_size(response, max_size)
        try:
            # json.loads doesn't allow bytes in Python 3.5
            json_body = json_decoder.decode(body.decode("UTF-8"))
        except Exception:
            logger.warning("Error parsing JSON from %s", uri)
            raise
        if not isinstance(json_body, dict):
            raise TypeError
        # Cast safety: json only permits strings as object keys, so `json_body`
        # must be Dict[str, Any] rather than Dict[Any, Any].
        return cast(JsonDict, json_body)

    async def post_json_get_nothing(
        self, uri: str, post_json: JsonDict, opts: Dict[str, Any]
    ) -> IResponse:
        """Make a POST request to an endpoint returning nothing

        :param uri: The URI to make a POST request to.

        :param post_json: A Python object that will be converted to a JSON
            string and POSTed to the given URI.

        :param opts: A dictionary of request options. Currently only opts.headers
            is supported.

        :return: a response from the remote server.
        """
        resp, _ = await self.post_json_maybe_get_json(uri, post_json, opts)
        return resp

    async def post_json_maybe_get_json(
        self,
        uri: str,
        post_json: Dict[str, Any],
        opts: Dict[str, Any],
        max_size: Optional[int] = None,
    ) -> Tuple[IResponse, Optional[JsonDict]]:
        """Make a POST request to an endpoint that might be returning JSON and parse
        result

        :param uri: The URI to make a POST request to.

        :param post_json: A Python object that will be converted to a JSON
            string and POSTed to the given URI.

        :param opts: A dictionary of request options. Currently only opts.headers
            is supported.

        :param max_size: The maximum size (in bytes) to allow as a response.

        :return: a response from the remote server, and its decoded JSON body if any (None
            otherwise).
        """
        json_bytes = json.dumps(post_json).encode("utf8")

        headers = opts.get(
            "headers",
            Headers(
                {
                    b"Content-Type": [b"application/json"],
                }
            ),
        )

        logger.debug("HTTP POST %s -> %s", json_bytes, uri)

        response = await self.agent.request(
            b"POST",
            uri.encode("utf8"),
            headers,
            bodyProducer=FileBodyProducer(BytesIO(json_bytes)),
        )

        # Ensure the body object is read otherwise we'll leak HTTP connections
        # as per
        # https://twistedmatrix.com/documents/current/web/howto/client.html
        json_body = None
        try:
            # TODO Will this cause the server to think the request was a failure?
            body = await read_body_with_max_size(response, max_size)
            json_body = json_decoder.decode(body.decode("UTF-8"))
        except Exception:
            # We might get an exception here because the body exceeds the max_size, or it
            # isn't valid JSON. In both cases, we don't care about it.
            pass

        return response, json_body


class SimpleHttpClient(HTTPClient[Agent]):
    """A simple, no-frills HTTP client based on the class of the same name
    from Synapse.
    """

    def __init__(self, sydent: "Sydent") -> None:
        self.sydent = sydent
        # The default endpoint factory in Twisted 14.0.0 (which we require) uses the
        # BrowserLikePolicyForHTTPS context factory which will do regular cert validation
        # 'like a browser'
        self.agent = Agent(
            BlacklistingReactorWrapper(
                reactor=self.sydent.reactor,
                ip_whitelist=sydent.config.general.ip_whitelist,
                ip_blacklist=sydent.config.general.ip_blacklist,
            ),
            connectTimeout=15,
        )


class FederationHttpClient(HTTPClient[MatrixFederationAgent]):
    """HTTP client for federation requests to homeservers. Uses a
    MatrixFederationAgent.
    """

    def __init__(self, sydent: "Sydent") -> None:
        self.sydent = sydent
        self.agent = MatrixFederationAgent(
            # Type-safety: I don't have a good way of expressing that
            # the reactor is IReactorTCP, IReactorTime and
            # IReactorPluggableNameResolver all at once. But it is, because
            # it wraps the sydent reactor.
            # TODO: can we introduce a SydentReactor type like SynapseReactor?
            BlacklistingReactorWrapper(  # type: ignore[arg-type]
                reactor=self.sydent.reactor,
                ip_whitelist=sydent.config.general.ip_whitelist,
                ip_blacklist=sydent.config.general.ip_blacklist,
            ),
            ClientTLSOptionsFactory(sydent.config.http.verify_federation_certs)
            if sydent.use_tls_for_federation
            else None,
        )



# Copyright 2014 OpenMarket Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
from typing import TYPE_CHECKING, List, cast

import twisted.python.log
from OpenSSL.crypto import X509
from twisted.internet.interfaces import ISSLTransport
from twisted.web.server import Request

from sydent.db.hashing_metadata import HashingMetadataStore
from sydent.db.peers import PeerStore
from sydent.db.threepid_associations import GlobalAssociationStore, SignedAssociations
from sydent.http.servlets import MatrixRestError, SydentResource, jsonwrap
from sydent.threepid import threePidAssocFromDict
from sydent.types import JsonDict
from sydent.util import json_decoder
from sydent.util.hash import sha256_and_url_safe_base64
from sydent.util.stringutils import normalise_address

if TYPE_CHECKING:
    from sydent.sydent import Sydent

logger = logging.getLogger(__name__)


class ReplicationPushServlet(SydentResource):
    def __init__(self, sydent: "Sydent") -> None:
        super().__init__()
        self.sydent = sydent
        self.hashing_store = HashingMetadataStore(sydent)

    @jsonwrap
    def render_POST(self, request: Request) -> JsonDict:
        # Cast safety: This request has an ISSLTransport because this servlet
        # is a resource under the ReplicationHttpsServer and nowhere else.
        request.transport = cast(ISSLTransport, request.transport)
        peerCert = cast(X509, request.transport.getPeerCertificate())
        peerCertCn = peerCert.get_subject().commonName

        peerStore = PeerStore(self.sydent)

        peer = peerStore.getPeerByName(peerCertCn)

        if not peer:
            logger.warning(
                "Got connection from %s but no peer found by that name", peerCertCn
            )
            raise MatrixRestError(
                403, "M_UNKNOWN_PEER", "This peer is not known to this server"
            )

        logger.info("Push connection made from peer %s", peer.servername)

        if (
            not request.requestHeaders.hasHeader("Content-Type")
            # Type safety: the hasHeader call returned True, so getRawHeaders()
            # returns a nonempty list.
            or request.requestHeaders.getRawHeaders("Content-Type")[0]  # type: ignore[index]
            != "application/json"
        ):
            logger.warning(
                "Peer %s made push connection with non-JSON content (type: %s)",
                peer.servername,
                # Type safety: the hasHeader call returned True, so getRawHeaders()
                # returns a nonempty list.
                request.requestHeaders.getRawHeaders("Content-Type")[0],  # type: ignore[index]
            )
            raise MatrixRestError(400, "M_NOT_JSON", "This endpoint expects JSON")

        try:
            # json.loads doesn't allow bytes in Python 3.5
            inJson = json_decoder.decode(request.content.read().decode("UTF-8"))
        except ValueError:
            logger.warning(
                "Peer %s made push connection with malformed JSON", peer.servername
            )
            raise MatrixRestError(400, "M_BAD_JSON", "Malformed JSON")

        if "sgAssocs" not in inJson:
            logger.warning(
                "Peer %s made push connection with no 'sgAssocs' key in JSON",
                peer.servername,
            )
            raise MatrixRestError(400, "M_BAD_JSON", 'No "sgAssocs" key in JSON')

        failedIds: List[int] = []

        globalAssocsStore = GlobalAssociationStore(self.sydent)

        # Ensure items are pulled out of the dictionary in order of origin_id.
        sg_assocs_raw: SignedAssociations = inJson.get("sgAssocs", {})
        sg_assocs = sorted(sg_assocs_raw.items(), key=lambda k: int(k[0]))

        for originId, sgAssoc in sg_assocs:
            try:
                peer.verifySignedAssociation(sgAssoc)
                logger.debug(
                    "Signed association from %s with origin ID %s verified",
                    peer.servername,
                    originId,
                )

                # Don't bother adding if one has already failed: we add all of them or none so
                # we're only going to roll back the transaction anyway (but we continue to try
                # & verify the rest so we can give a complete list of the ones that don't
                # verify)
                if len(failedIds) > 0:
                    continue

                assocObj = threePidAssocFromDict(sgAssoc)

                # ensure we are casefolding email addresses before hashing/storing
                assocObj.address = normalise_address(assocObj.address, assocObj.medium)

                if assocObj.mxid is not None:
                    # Calculate the lookup hash with our own pepper for this association
                    pepper = self.hashing_store.get_lookup_pepper()
                    assert pepper is not None
                    str_to_hash = " ".join(
                        [assocObj.address, assocObj.medium, pepper],
                    )
                    assocObj.lookup_hash = sha256_and_url_safe_base64(str_to_hash)

                    # Add this association
                    globalAssocsStore.addAssociation(
                        assocObj,
                        json.dumps(sgAssoc),
                        peer.servername,
                        originId,
                        commit=False,
                    )
                else:
                    logger.info(
                        "Incoming deletion: removing associations for %s / %s",
                        assocObj.medium,
                        assocObj.address,
                    )
                    globalAssocsStore.removeAssociation(
                        assocObj.medium, assocObj.address
                    )
                logger.info(
                    "Stored association origin ID %s from %s", originId, peer.servername
                )
            except Exception:
                failedIds.append(originId)
                logger.warning(
                    "Failed to verify signed association from %s with origin ID %s",
                    peer.servername,
                    originId,
                )
                twisted.python.log.err()

        if len(failedIds) > 0:
            self.sydent.db.rollback()
            request.setResponseCode(400)
            return {
                "errcode": "M_VERIFICATION_FAILED",
                "error": "Verification failed for one or more associations",
                "failed_ids": failedIds,
            }
        else:
            self.sydent.db.commit()
            return {"success": True}



# Copyright 2014 OpenMarket Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING

from twisted.web.server import Request
from unpaddedbase64 import encode_base64

from sydent.db.invite_tokens import JoinTokenStore
from sydent.http.servlets import SydentResource, get_args, jsonwrap
from sydent.types import JsonDict

if TYPE_CHECKING:
    from sydent.sydent import Sydent


class Ed25519Servlet(SydentResource):
    isLeaf = True

    def __init__(self, syd: "Sydent") -> None:
        super().__init__()
        self.sydent = syd

    @jsonwrap
    def render_GET(self, request: Request) -> JsonDict:
        pubKey = self.sydent.keyring.ed25519.verify_key
        pubKeyBase64 = encode_base64(pubKey.encode())

        return {"public_key": pubKeyBase64}


class PubkeyIsValidServlet(SydentResource):
    isLeaf = True

    def __init__(self, syd: "Sydent") -> None:
        super().__init__()
        self.sydent = syd

    @jsonwrap
    def render_GET(self, request: Request) -> JsonDict:
        args = get_args(request, ("public_key",))

        pubKey = self.sydent.keyring.ed25519.verify_key
        pubKeyBase64 = encode_base64(pubKey.encode())

        return {"valid": args["public_key"] == pubKeyBase64}


class EphemeralPubkeyIsValidServlet(SydentResource):
    isLeaf = True

    def __init__(self, syd: "Sydent") -> None:
        super().__init__()
        self.joinTokenStore = JoinTokenStore(syd)

    @jsonwrap
    def render_GET(self, request: Request) -> JsonDict:
        args = get_args(request, ("public_key",))
        publicKey = args["public_key"]

        return {
            "valid": self.joinTokenStore.validateEphemeralPublicKey(publicKey),
        }



# Copyright 2015 OpenMarket Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import random
import re
import string
from email.header import Header
from http import HTTPStatus
from typing import TYPE_CHECKING

import nacl.signing
from twisted.web.server import Request
from unpaddedbase64 import encode_base64

from sydent.db.invite_tokens import JoinTokenStore
from sydent.db.threepid_associations import GlobalAssociationStore
from sydent.http.auth import authV2
from sydent.http.servlets import (
    MatrixRestError,
    SydentResource,
    get_args,
    jsonwrap,
    send_cors,
)
from sydent.types import JsonDict
from sydent.util.emailutils import EmailAddressException, sendEmail
from sydent.util.stringutils import MAX_EMAIL_ADDRESS_LENGTH, normalise_address

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from sydent.sydent import Sydent


class StoreInviteServlet(SydentResource):
    def __init__(self, syd: "Sydent", require_auth: bool = False) -> None:
        super().__init__()
        self.sydent = syd
        self.random = random.SystemRandom()
        self.require_auth = require_auth

    @jsonwrap
    def render_POST(self, request: Request) -> JsonDict:
        send_cors(request)

        args = get_args(
            request,
            (
                "medium",
                "address",
                "room_id",
                "sender",
            ),
        )
        medium = args["medium"]
        address = args["address"]
        roomId = args["room_id"]
        sender = args["sender"]

        # ensure we are casefolding email address before storing
        normalised_address = normalise_address(address, medium)

        verified_sender = None
        if self.require_auth:
            account = authV2(self.sydent, request)
            verified_sender = sender
            if account.userId != sender:
                raise MatrixRestError(403, "M_UNAUTHORIZED", "'sender' doesn't match")

        logger.info(
            "Store invite request from %s to %s, in %s", sender, address, roomId
        )

        sender_homeserver = sender.split(":", 1)[1]
        if (
            roomId in self.sydent.config.email.third_party_invite_room_blocklist
            or sender_homeserver
            in self.sydent.config.email.third_party_invite_homeserver_blocklist
        ):
            raise MatrixRestError(403, "M_UNAUTHORIZED", "Invite not allowed")

        self.sydent.email_sender_ratelimiter.ratelimit(
            sender, "Limit exceeded for this sender"
        )

        globalAssocStore = GlobalAssociationStore(self.sydent)
        mxid = globalAssocStore.getMxid(medium, normalised_address)
        if mxid:
            request.setResponseCode(400)
            return {
                "errcode": "M_THREEPID_IN_USE",
                "error": "Binding already known",
                "mxid": mxid,
            }

        if medium != "email":
            request.setResponseCode(400)
            return {
                "errcode": "M_UNRECOGNIZED",
                "error": "Didn't understand medium '%s'" % (medium,),
            }

        if not (0 < len(address) <= MAX_EMAIL_ADDRESS_LENGTH):
            request.setResponseCode(400)
            return {"errcode": "M_INVALID_PARAM", "error": "Invalid email provided"}

        token = self._randomString(128)

        tokenStore = JoinTokenStore(self.sydent)

        ephemeralPrivateKey = nacl.signing.SigningKey.generate()
        ephemeralPublicKey = ephemeralPrivateKey.verify_key

        ephemeralPrivateKeyBase64 = encode_base64(ephemeralPrivateKey.encode(), True)
        ephemeralPublicKeyBase64 = encode_base64(ephemeralPublicKey.encode(), True)

        tokenStore.storeEphemeralPublicKey(ephemeralPublicKeyBase64)
        tokenStore.storeToken(medium, normalised_address, roomId, sender, token)

        # Variables to substitute in the template.
        substitutions = {}
        # Include all arguments sent via the request.
        for k, v in args.items():
            if isinstance(v, str):
                substitutions[k] = v
        substitutions["token"] = token

        for keyword in self.sydent.config.email.third_party_invite_keyword_blocklist:
            for (key, value) in args.items():
                # make sure the blocklist doesn't stomp on web_client_location url
                if key == "org.matrix.web_client_location":
                    value = re.sub(r"^https?://", "", value)
                if keyword in value.casefold():
                    logger.info(
                        "Denying invites as %r appears in arg %r: %r",
                        keyword,
                        key,
                        value,
                    )
                    raise MatrixRestError(403, "M_UNAUTHORIZED", "Invite not allowed")

        # Substitutions that the template requires, but are optional to provide
        # to the API.
        extra_substitutions = [
            "sender_display_name",
            "token",
            "room_name",
            "bracketed_room_name",
            "room_avatar_url",
            "sender_avatar_url",
            "guest_user_id",
            "guest_access_token",
            "room_type",
        ]
        for k in extra_substitutions:
            substitutions.setdefault(k, "")

        if "sender_display_name" in substitutions:
            if len(substitutions["sender_display_name"]) > 20:
                substitutions["sender_display_name"] = (
                    substitutions["sender_display_name"][:19] + "…"
                )

        substitutions["bracketed_verified_sender"] = ""
        if verified_sender:
            substitutions["bracketed_verified_sender"] = "(%s) " % (verified_sender,)

        substitutions["ephemeral_private_key"] = ephemeralPrivateKeyBase64
        if substitutions["room_name"] != "":
            if len(substitutions["room_name"]) > 30:
                substitutions["room_name"] = substitutions["room_name"][:25] + "…"

            substitutions["bracketed_room_name"] = "(%s) " % substitutions["room_name"]

        substitutions[
            "web_client_location"
        ] = self.sydent.config.email.default_web_client_location
        if "org.matrix.web_client_location" in substitutions:
            substitutions["web_client_location"] = substitutions[
                "org.matrix.web_client_location"
            ]

        if substitutions["room_type"] == "m.space":
            subject = self.sydent.config.email.invite_subject_space % substitutions
        else:
            subject = self.sydent.config.email.invite_subject % substitutions

        substitutions["subject_header_value"] = Header(subject, "utf8").encode()

        brand = self.sydent.brand_from_request(request)

        # self.sydent.config.email.invite_template is deprecated
        if self.sydent.config.email.invite_template is None:
            templateFile = self.sydent.get_branded_template(
                brand,
                "invite_template.eml",
            )
        else:
            templateFile = self.sydent.config.email.invite_template

        try:
            sendEmail(self.sydent, templateFile, normalised_address, substitutions)
        except EmailAddressException:
            request.setResponseCode(HTTPStatus.BAD_REQUEST)
            return {"errcode": "M_INVALID_EMAIL", "error": "Invalid email address"}

        pubKey = self.sydent.keyring.ed25519.verify_key
        pubKeyBase64 = encode_base64(pubKey.encode())

        baseUrl = "%s/_matrix/identity/api/v1" % (
            self.sydent.config.http.server_http_url_base,
        )

        keysToReturn = []
        keysToReturn.append(
            {
                "public_key": pubKeyBase64,
                "key_validity_url": baseUrl + "/pubkey/isvalid",
            }
        )
        keysToReturn.append(
            {
                "public_key": ephemeralPublicKeyBase64,
                "key_validity_url": baseUrl + "/pubkey/ephemeral/isvalid",
            }
        )

        resp = {
            "token": token,
            "public_key": pubKeyBase64,
            "public_keys": keysToReturn,
            "display_name": self.redact_email_address(address),
        }

        return resp

    def redact_email_address(self, address: str) -> str:
        """
        Redacts the content of a 3PID address. Redacts both the email's username and
        domain independently.

        :param address: The address to redact.

        :return: The redacted address.
        """
        # Extract strings from the address
        username, domain = address.split("@", 1)

        # Obfuscate strings
        redacted_username = self._redact(
            username, self.sydent.config.email.username_obfuscate_characters
        )
        redacted_domain = self._redact(
            domain, self.sydent.config.email.domain_obfuscate_characters
        )

        return redacted_username + "@" + redacted_domain

    def _redact(self, s: str, characters_to_reveal: int) -> str:
        """
        Redacts the content of a string, using a given amount of characters to reveal.
        If the string is shorter than the given threshold, redact it based on length.

        :param s: The string to redact.

        :param characters_to_reveal: How many characters of the string to leave before
            the '...'

        :return: The redacted string.
        """
        # If the string is shorter than the defined threshold, redact based on length
        if len(s) <= characters_to_reveal:
            if len(s) > 5:
                return s[:3] + "..."
            if len(s) > 1:
                return s[0] + "..."
            return "..."

        # Otherwise truncate it and add an ellipses
        return s[:characters_to_reveal] + "..."

    def _randomString(self, length: int) -> str:
        """
        Generate a random string of the given length.

        :param length: The length of the string to generate.

        :return: The generated string.
        """
        return "".join(self.random.choice(string.ascii_letters) for _ in range(length))



# Copyright 2019 New Vector Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
from typing import Callable

from OpenSSL import SSL
from twisted.internet import ssl
from twisted.internet.abstract import isIPAddress, isIPv6Address
from twisted.internet.interfaces import IOpenSSLClientConnectionCreator
from twisted.protocols.tls import TLSMemoryBIOProtocol
from twisted.python.failure import Failure
from zope.interface import implementer

logger = logging.getLogger(__name__)

F = Callable[[SSL.Connection, int, int], None]


def _tolerateErrors(wrapped: F) -> F:
    """
    Wrap up an info_callback for pyOpenSSL so that if something goes wrong
    the error is immediately logged and the connection is dropped if possible.
    This is a copy of twisted.internet._sslverify._tolerateErrors. For
    documentation, see the twisted documentation.
    """

    def infoCallback(connection: SSL.Connection, where: int, ret: int) -> None:
        try:
            return wrapped(connection, where, ret)
        except BaseException:
            f = Failure()
            logger.exception("Error during info_callback")
            connection.get_app_data().failVerification(f)

    return infoCallback


def _idnaBytes(text: str) -> bytes:
    """
    Convert some text typed by a human into some ASCII bytes. This is a
    copy of twisted.internet._idna._idnaBytes. For documentation, see the
    twisted documentation.
    """
    try:
        import idna
    except ImportError:
        return text.encode("idna")
    else:
        return idna.encode(text)


@implementer(IOpenSSLClientConnectionCreator)
class ClientTLSOptions:
    """
    Client creator for TLS without certificate identity verification. This is a
    copy of twisted.internet._sslverify.ClientTLSOptions with the identity
    verification left out. For documentation, see the twisted documentation.
    """

    def __init__(self, hostname: str, ctx: SSL.Context):
        self._ctx = ctx

        if isIPAddress(hostname) or isIPv6Address(hostname):
            self._hostnameBytes = hostname.encode("ascii")
            self._sendSNI = False
        else:
            self._hostnameBytes = _idnaBytes(hostname)
            self._sendSNI = True

        ctx.set_info_callback(_tolerateErrors(self._identityVerifyingInfoCallback))

    def clientConnectionForTLS(
        self, tlsProtocol: TLSMemoryBIOProtocol
    ) -> SSL.Connection:
        context = self._ctx
        connection = SSL.Connection(context, None)
        connection.set_app_data(tlsProtocol)
        return connection

    def _identityVerifyingInfoCallback(
        self, connection: SSL.Connection, where: int, ret: int
    ) -> None:
        # Literal IPv4 and IPv6 addresses are not permitted
        # as host names according to the RFCs
        if where & SSL.SSL_CB_HANDSHAKE_START and self._sendSNI:
            connection.set_tlsext_host_name(self._hostnameBytes)


class ClientTLSOptionsFactory:
    """Factory for Twisted ClientTLSOptions that are used to make connections
    to remote servers for federation."""

    def __init__(self, verify_requests: bool):
        if verify_requests:
            self._options = ssl.CertificateOptions(trustRoot=ssl.platformTrust())
        else:
            self._options = ssl.CertificateOptions()

    def get_options(self, host: str) -> ClientTLSOptions:
        # Use _makeContext so that we get a fresh OpenSSL CTX each time.
        return ClientTLSOptions(host, self._options._makeContext())
